import asyncio
import logging
from pathlib import Path
import time

import numpy as np
import torch
from torch import Tensor
from typing import List, Tuple, Union
import ipywidgets as widgets
from IPython.display import display
from huggingface_hub import hf_hub_download

from reflectorch.data_generation.priors import Params, BasicParams, ExpUniformSubPriorSampler, UniformSubPriorParams
from reflectorch.data_generation.priors.parametric_models import NuisanceParamsWrapper
from reflectorch.data_generation.q_generator import ConstantQ, VariableQ, MaskedVariableQ
from reflectorch.data_generation.utils import get_density_profiles, get_param_labels
from reflectorch.inference.preprocess_exp.interpolation import interp_reflectivity
from reflectorch.paths import CONFIG_DIR, ROOT_DIR, SAVED_MODELS_DIR
from reflectorch.runs.utils import (
    get_trainer_by_name, train_from_config
)
from reflectorch.runs.config import load_config
from reflectorch.ml.trainers import PointEstimatorTrainer
from reflectorch.data_generation.likelihoods import LogLikelihood

from reflectorch.inference.preprocess_exp import StandardPreprocessing
from reflectorch.inference.scipy_fitter import standard_refl_fit, refl_fit, get_fit_with_growth
from reflectorch.inference.sampler_solution import simple_sampler_solution, get_best_mse_param
from reflectorch.inference.record_time import print_time
from reflectorch.inference.plotting import plot_reflectivity, plot_prediction_results, print_prediction_results
from reflectorch.utils import get_filtering_mask, to_t

class EasyInferenceModel(object):
    """Facilitates the inference process using pretrained models
    
    Args:
        config_name (str, optional): the name of the configuration file used to initialize the model (either with or without the '.yaml' extension). Defaults to None.
        model_name (str, optional): the name of the file containing the weights of the model (either with or without the '.pt' extension), only required if different than: `'model_' + config_name + '.pt'`. Defaults to None 
        root_dir (str, optional): path to root directory containing the 'configs' and 'saved_models' subdirectories, if different from the package root directory (ROOT_DIR). Defaults to None.
        weights_format (str, optional): format (extension) of the weights file, either 'pt' or 'safetensors'. Defaults to 'safetensors'.
        repo_id (str, optional): the id of the Huggingface repository from which the configuration files and model weights should be downloaded automatically if not found locally (in the 'configs' and 'saved_models' subdirectories of the root directory). Defaults to 'valentinsingularity/reflectivity'.
        trainer (PointEstimatorTrainer, optional): if provided, this trainer instance is used directly instead of being initialized from the configuration file. Defaults to None.
        device (str, optional): the Pytorch device ('cuda' or 'cpu'). Defaults to 'cuda'.
    """
    def __init__(self, config_name: str = None, model_name: str = None, root_dir:str = None, weights_format: str = 'safetensors',
                 repo_id: str = 'valentinsingularity/reflectivity', trainer: PointEstimatorTrainer = None, device='cuda'):
        self.config_name = config_name
        self.model_name = model_name
        self.root_dir = root_dir
        self.weights_format = weights_format
        self.repo_id = repo_id
        self.trainer = trainer
        self.device = device

        if trainer is None and self.config_name is not None:
            self.load_model(self.config_name, self.model_name, self.root_dir)

        self.prediction_result = None

    def load_model(self, config_name: str, model_name: str, root_dir: str) -> None:
        """Loads a model for inference

        Args:
            config_name (str): the name of the configuration file used to initialize the model (either with or without the '.yaml' extension).
            model_name (str): the name of the file containing the weights of the model (either with or without the '.pt' or '.safetensors'  extension), only required if different than: `'model_' + config_name + extension`.
            root_dir (str): path to root directory containing the 'configs' and 'saved_models' subdirectories, if different from the package root directory (ROOT_DIR).
        """
        if self.config_name == config_name and self.trainer is not None:
            return
        
        if not config_name.endswith('.yaml'):
            config_name_no_extension = config_name
            self.config_name = config_name_no_extension + '.yaml'
        else:
            config_name_no_extension = config_name[:-5]
            self.config_name = config_name
        
        self.config_dir = Path(root_dir) / 'configs' if root_dir else CONFIG_DIR
        weights_extension = '.' + self.weights_format
        self.model_name = model_name or 'model_' + config_name_no_extension + weights_extension
        if not self.model_name.endswith(weights_extension):
            self.model_name += weights_extension
        self.model_dir = Path(root_dir) / 'saved_models' if root_dir else SAVED_MODELS_DIR

        config_path = Path(self.config_dir) / self.config_name
        if config_path.exists():
            print(f"Configuration file `{config_path}` found locally.")
        else:
            print(f"Configuration file `{config_path}` not found locally.")
            if self.repo_id is None:
                raise ValueError("repo_id must be provided to download files from Huggingface.")
            print("Downloading from Huggingface...")
            hf_hub_download(repo_id=self.repo_id, subfolder='configs', filename=self.config_name, local_dir=config_path.parents[1])

        model_path = Path(self.model_dir) / self.model_name
        if model_path.exists():
            print(f"Weights file `{model_path}` found locally.")
        else:
            print(f"Weights file `{model_path}` not found locally.")
            if self.repo_id is None:
                raise ValueError("repo_id must be provided to download files from Huggingface.")
            print("Downloading from Huggingface...")
            hf_hub_download(repo_id=self.repo_id, subfolder='saved_models', filename=self.model_name, local_dir=model_path.parents[1])

        self.trainer = get_trainer_by_name(config_name=config_name, config_dir=self.config_dir, model_path=model_path, load_weights=True, inference_device = self.device)
        self.trainer.model.eval()
        
        param_model = self.trainer.loader.prior_sampler.param_model
        param_model_name = param_model.base_model.NAME if isinstance(param_model, NuisanceParamsWrapper) else param_model.NAME
        print(f'The model corresponds to a `{param_model_name}` parameterization with {self.trainer.loader.prior_sampler.max_num_layers} layers ({self.trainer.loader.prior_sampler.param_dim} predicted parameters)')
        print("Parameter types and total ranges:")
        for param, range_ in self.trainer.loader.prior_sampler.param_ranges.items():
            print(f"- {param}: {range_}")
        print("Allowed widths of the prior bound intervals (max-min):")
        for param, range_ in self.trainer.loader.prior_sampler.bound_width_ranges.items():
            print(f"- {param}: {range_}")

        if isinstance(self.trainer.loader.q_generator, ConstantQ):
            q_min = self.trainer.loader.q_generator.q[0].item()
            q_max = self.trainer.loader.q_generator.q[-1].item()
            n_q = self.trainer.loader.q_generator.q.shape[0]
            print(f'The model was trained on curves discretized at {n_q} uniform points between q_min={q_min} and q_max={q_max}')
        elif isinstance(self.trainer.loader.q_generator, VariableQ):
            q_min_range = self.trainer.loader.q_generator.q_min_range
            q_max_range = self.trainer.loader.q_generator.q_max_range
            n_q_range = self.trainer.loader.q_generator.n_q_range
            if n_q_range[0] == n_q_range[1]:
                n_q_fixed = n_q_range[0]
                print(f'The model was trained on curves discretized at exactly {n_q_fixed} uniform points, '
                    f'between q_min in [{q_min_range[0]}, {q_min_range[1]}] and q_max in [{q_max_range[0]}, {q_max_range[1]}]')
            else:
                print(f'The model was trained on curves discretized at a number between {n_q_range[0]} and {n_q_range[1]} '
                    f'of uniform points between q_min in [{q_min_range[0]}, {q_min_range[1]}] and q_max in [{q_max_range[0]}, {q_max_range[1]}]')
        
        if self.trainer.loader.smearing is not None:
            q_res_min = self.trainer.loader.smearing.sigma_min
            q_res_max = self.trainer.loader.smearing.sigma_max
            if self.trainer.loader.smearing.constant_dq == False:
                print(f"The model was trained with linear resolution smearing (dq/q) in the range [{q_res_min}, {q_res_max}]")
            elif self.trainer.loader.smearing.constant_dq == True:
                print(f"The model was trained with constant resolution smearing in the range [{q_res_min}, {q_res_max}]")

        additional_inputs = ["prior bounds"]
        if self.trainer.train_with_q_input:
            additional_inputs.append("q values")
        if self.trainer.condition_on_q_resolutions:
            additional_inputs.append("the resolution dq/q")
        if additional_inputs:
            inputs_str = ", ".join(additional_inputs)
            print(f"The following quantities are additional inputs to the network: {inputs_str}.")

    def preprocess_and_predict(self, 
                            reflectivity_curve: Union[np.ndarray, Tensor], 
                            q_values: Union[np.ndarray, Tensor] = None, 
                            prior_bounds: Union[np.ndarray, List[Tuple]] = None, 
                            sigmas: Union[np.ndarray, Tensor] = None,
                            q_resolution: float = None,
                            ambient_sld: float = None,
                            clip_prediction: bool = True, 
                            polish_prediction: bool = False,
                            polishing_method: str = 'trf',
                            polishing_kwargs_reflectivity: dict = None,
                            use_sigmas_for_polishing: bool = False,
                            polishing_max_nfev: int = None,
                            fit_growth: bool = False, 
                            max_d_change: float = 5.,
                            calc_pred_curve: bool = True,
                            calc_pred_sld_profile: bool = False,
                            calc_polished_sld_profile: bool = False,
                            sld_profile_padding_left: float = 0.2,
                            sld_profile_padding_right: float = 1.1,
                            kwargs_param_labels: dict = {},
                            
                            truncate_index_left: int = None,
                            truncate_index_right: int = None,
                            enable_error_bars_filtering: bool = True,
                            filter_threshold=0.3,
                            filter_remove_singles=True,
                            filter_remove_consecutives=True,
                            filter_consecutive=3,
                            filter_q_start_trunc=0.1,
                            ):
        
        ## Preprocess the data for inference (remove negative intensities, truncation, filer out points with high error bars)
        (q_values, reflectivity_curve, sigmas, q_resolution, 
         q_values_original, reflectivity_curve_original, sigmas_original, q_resolution_original) = self._preprocess_input_data(
            reflectivity_curve=reflectivity_curve,
            q_values=q_values,
            sigmas=sigmas,
            q_resolution=q_resolution,
            truncate_index_left=truncate_index_left,
            truncate_index_right=truncate_index_right,
            enable_error_bars_filtering=enable_error_bars_filtering,
            filter_threshold=filter_threshold,
            filter_remove_singles=filter_remove_singles,
            filter_remove_consecutives=filter_remove_consecutives,
            filter_consecutive=filter_consecutive,
            filter_q_start_trunc=filter_q_start_trunc,
        )

        ### Interpolate the experimental data if needed by the embedding network
        interp_data = self.interpolate_data_to_model_q(
            q_exp=q_values,
            refl_exp=reflectivity_curve,
            sigmas_exp=sigmas,
            q_res_exp=q_resolution,
            as_dict=True
        )

        q_model = interp_data["q_model"]
        reflectivity_curve_interp = interp_data["reflectivity"]
        sigmas_interp = interp_data.get("sigmas")
        q_resolution_interp = interp_data.get("q_resolution")
        key_padding_mask = interp_data.get("key_padding_mask")
        
        ### Make the prediction
        prediction_dict = self.predict(
            reflectivity_curve=reflectivity_curve_interp,
            q_values=q_model,
            sigmas=sigmas_interp,
            q_resolution=q_resolution_interp,
            key_padding_mask=key_padding_mask,
            prior_bounds=prior_bounds,
            ambient_sld=ambient_sld,
            clip_prediction=clip_prediction,
            polish_prediction=False, ###do the polishing outside the predict method on the full data
            supress_sld_amb_back_shift=True, ###do not shift back the slds by the ambient yet
            calc_pred_curve=calc_pred_curve,
            calc_pred_sld_profile=calc_pred_sld_profile,
            sld_profile_padding_left=sld_profile_padding_left,
            sld_profile_padding_right=sld_profile_padding_right,
            kwargs_param_labels=kwargs_param_labels,
        )
        
        ### Save interpolated data
        prediction_dict['q_model'] = q_model
        prediction_dict['reflectivity_curve_interp'] = reflectivity_curve_interp
        if q_resolution_interp is not None:
            prediction_dict['q_resolution_interp'] = q_resolution_interp 
        if sigmas_interp is not None:
            prediction_dict['sigmas_interp'] = sigmas_interp
        if key_padding_mask is not None:
            prediction_dict['key_padding_mask'] = key_padding_mask

        ### Perform polishing on the original data
        if polish_prediction:
            polishing_kwargs = polishing_kwargs_reflectivity or {}
            polishing_kwargs.setdefault('dq', q_resolution_original)
            
            prior_bounds = np.array(prior_bounds)
            if ambient_sld:
                sld_indices = self._shift_slds_by_ambient(prior_bounds, ambient_sld)

            polished_dict = self._polish_prediction(
                q=q_values_original,
                curve=reflectivity_curve_original,
                predicted_params=prediction_dict['predicted_params_object'],
                priors=prior_bounds,
                ambient_sld_tensor=torch.atleast_2d(torch.as_tensor(ambient_sld)).to(self.device) if ambient_sld is not None else None,
                calc_polished_sld_profile=calc_polished_sld_profile,
                sld_x_axis=torch.from_numpy(prediction_dict['predicted_sld_xaxis']),
                polishing_kwargs_reflectivity = polishing_kwargs,
                error_bars=sigmas_original if use_sigmas_for_polishing else None,
                polishing_method=polishing_method,
                polishing_max_nfev=polishing_max_nfev,
                fit_growth=fit_growth,
                max_d_change=max_d_change,
            )

            prediction_dict.update(polished_dict)
            if fit_growth and "polished_params_array" in prediction_dict:
                prediction_dict["param_names"].append("max_d_change")
        
        ### Shift back the slds for nonzero ambient
        if ambient_sld:
            self._restore_slds_after_ambient_shift(prediction_dict, sld_indices, ambient_sld)   

        return prediction_dict


    def predict(self, 
                reflectivity_curve: Union[np.ndarray, Tensor], 
                q_values: Union[np.ndarray, Tensor] = None, 
                prior_bounds: Union[np.ndarray, List[Tuple]] = None, 
                sigmas: Union[np.ndarray, Tensor] = None,
                key_padding_mask: Union[np.ndarray, Tensor] = None,
                q_resolution: float = None,
                ambient_sld: float = None,
                clip_prediction: bool = True, 
                polish_prediction: bool = False,
                polishing_method: str = 'trf',
                polishing_kwargs_reflectivity: dict = None,
                polishing_max_nfev: int = None,
                fit_growth: bool = False, 
                max_d_change: float = 5.,
                use_q_shift: bool = False, 
                calc_pred_curve: bool = True,
                calc_pred_sld_profile: bool = False,
                calc_polished_sld_profile: bool = False,
                sld_profile_padding_left: float = 0.2,
                sld_profile_padding_right: float = 1.1,
                supress_sld_amb_back_shift: bool = False,
                kwargs_param_labels: dict = {},
                ):
        """Predict the thin film parameters

        Args:
            reflectivity_curve (Union[np.ndarray, Tensor]): The reflectivity curve (which has been already preprocessed, normalized and interpolated).
            q_values (Union[np.ndarray, Tensor], optional): The momentum transfer (q) values for the reflectivity curve (in units of inverse angstroms).
            prior_bounds (Union[np.ndarray, List[Tuple]], optional): the prior bounds for the thin film parameters.
            clip_prediction (bool, optional): If ``True``, the values of the predicted parameters are clipped to not be outside the interval set by the prior bounds. Defaults to False.
            polish_prediction (bool, optional): If ``True``, the neural network predictions are further polished using a simple least mean squares (LMS) fit. Only for the standard box-model parameterization. Defaults to False.
            fit_growth (bool, optional): If ``True``, an additional parameters is introduced during the LMS polishing to account for the change in the thickness of the upper layer during the in-situ measurement of the reflectivity curve (a linear growth is assumed). Defaults to False.
            max_d_change (float): The maximum possible change in the thickness of the upper layer during the in-situ measurement, relevant when polish_prediction and fit_growth are True. Defaults to 5. 
            use_q_shift: If ``True``, the prediction is performed for a batch of slightly shifted versions of the input curve and the best result is returned, which is meant to mitigate the influence of imperfect sample alignment, as introduced in Greco et al. (only for models with fixed q-discretization). Defaults to False.
            calc_pred_curve (bool, optional): Whether to calculate the curve corresponding to the predicted parameters. Defaults to True.
            calc_pred_sld_profile (bool, optional): Whether to calculate the SLD profile corresponding to the predicted parameters. Defaults to False.

        Returns:
            dict: dictionary containing the predictions
        """

        scaled_curve = self._scale_curve(reflectivity_curve)
        prior_bounds = np.array(prior_bounds)

        if ambient_sld:
            sld_indices = self._shift_slds_by_ambient(prior_bounds, ambient_sld)

        scaled_prior_bounds = self._scale_prior_bounds(prior_bounds)

        if isinstance(self.trainer.loader.q_generator, ConstantQ):
            q_values = self.trainer.loader.q_generator.q
        else:
            q_values = torch.atleast_2d(to_t(q_values)).to(scaled_curve)

        scaled_q_values = self.trainer.loader.q_generator.scale_q(q_values).to(torch.float32) if self.trainer.train_with_q_input else None
        
        if q_resolution is not None:
            q_resolution_tensor = torch.atleast_2d(torch.as_tensor(q_resolution)).to(scaled_curve)
            if isinstance(q_resolution, float):
                unscaled_q_resolutions = q_resolution_tensor
            else:
                unscaled_q_resolutions = (q_resolution_tensor / q_values).nanmean(dim=-1, keepdim=True) ##when q_values is padded with 0s, there will be nan at the padded positions
            scaled_q_resolutions = self.trainer.loader.smearing.scale_resolutions(unscaled_q_resolutions) if self.trainer.condition_on_q_resolutions else None
            scaled_conditioning_params = scaled_q_resolutions
            if polishing_kwargs_reflectivity is None:
                polishing_kwargs_reflectivity = {'dq': q_resolution}
        else:
            q_resolution_tensor = None
            scaled_conditioning_params = None

        if key_padding_mask is not None:
            key_padding_mask = torch.as_tensor(key_padding_mask, device=self.device)
            key_padding_mask = key_padding_mask.unsqueeze(0) if key_padding_mask.dim() == 1 else key_padding_mask

        if use_q_shift and not self.trainer.train_with_q_input:
            predicted_params = self._qshift_prediction(reflectivity_curve, scaled_prior_bounds, num = 1024, dq_coef = 1.)
        else:
            with torch.no_grad():
                self.trainer.model.eval()
    
                scaled_predicted_params = self.trainer.model(
                    curves=scaled_curve, 
                    bounds=scaled_prior_bounds, 
                    q_values=scaled_q_values,
                    conditioning_params = scaled_conditioning_params,
                    key_padding_mask = key_padding_mask,
                    unscaled_q_values = q_values,
                    )
                
                predicted_params = self.trainer.loader.prior_sampler.restore_params(torch.cat([scaled_predicted_params, scaled_prior_bounds], dim=-1))

        if clip_prediction:
            predicted_params = self.trainer.loader.prior_sampler.clamp_params(predicted_params)
        
        prediction_dict = {
            "predicted_params_object": predicted_params,
            "predicted_params_array": predicted_params.parameters.squeeze().cpu().numpy(),
            "param_names" : self.trainer.loader.prior_sampler.param_model.get_param_labels(**kwargs_param_labels)
        }
        
        key_padding_mask = None if key_padding_mask is None else key_padding_mask.squeeze().cpu().numpy()

        if calc_pred_curve:
            predicted_curve = predicted_params.reflectivity(q=q_values, dq=q_resolution_tensor).squeeze().cpu().numpy()
            prediction_dict[ "predicted_curve"] = predicted_curve if key_padding_mask is None else predicted_curve[key_padding_mask]
        
        ambient_sld_tensor = torch.atleast_2d(torch.as_tensor(ambient_sld, device=self.device)) if ambient_sld is not None else None
        if calc_pred_sld_profile: 
            predicted_sld_xaxis, predicted_sld_profile, _ = get_density_profiles(
                predicted_params.thicknesses, predicted_params.roughnesses, predicted_params.slds + (ambient_sld_tensor or 0), ambient_sld_tensor, 
                num=1024, padding_left=sld_profile_padding_left, padding_right=sld_profile_padding_right,
            ) 
            prediction_dict['predicted_sld_profile'] = predicted_sld_profile.squeeze().cpu().numpy()
            prediction_dict['predicted_sld_xaxis'] = predicted_sld_xaxis.squeeze().cpu().numpy()
        else:
            predicted_sld_xaxis = None
        
        refl_curve_polish = reflectivity_curve if key_padding_mask is None else reflectivity_curve[key_padding_mask]
        q_polish = q_values.squeeze().cpu().numpy() if key_padding_mask is None else q_values.squeeze().cpu().numpy()[key_padding_mask]
        prediction_dict['q_plot_pred'] = q_polish

        if polish_prediction:
            if ambient_sld_tensor:
                ambient_sld_tensor = ambient_sld_tensor.cpu()
            
            polished_dict = self._polish_prediction(
                q = q_polish, 
                curve = refl_curve_polish, 
                predicted_params = predicted_params, 
                priors = np.array(prior_bounds), 
                error_bars = sigmas,
                fit_growth = fit_growth,
                max_d_change = max_d_change, 
                calc_polished_curve = calc_pred_curve,
                calc_polished_sld_profile = calc_polished_sld_profile,
                ambient_sld_tensor=ambient_sld_tensor,
                sld_x_axis = predicted_sld_xaxis,
                polishing_method=polishing_method,
                polishing_max_nfev=polishing_max_nfev,
                polishing_kwargs_reflectivity=polishing_kwargs_reflectivity,
            )
            prediction_dict.update(polished_dict)

            if fit_growth and "polished_params_array" in prediction_dict:
                prediction_dict["param_names"].append("max_d_change")

        if ambient_sld and not supress_sld_amb_back_shift: #Note: the SLD shift will only be reflected in predicted_params_array but not in predicted_params_object; supress_sld_amb_back_shift is required for the 'preprocess_and_predict' method
            self._restore_slds_after_ambient_shift(prediction_dict, sld_indices, ambient_sld)   

        return prediction_dict
    
    def _polish_prediction(self,
                           q: np.ndarray,
                           curve: np.ndarray,
                           predicted_params: BasicParams,
                           priors: np.ndarray,
                           sld_x_axis,
                           ambient_sld_tensor: Tensor = None,
                           fit_growth: bool = False,
                           max_d_change: float = 5.,
                           calc_polished_curve: bool = True,
                           calc_polished_sld_profile: bool = False,
                           error_bars: np.ndarray = None,
                           polishing_method: str = 'trf',
                           polishing_max_nfev: int = None,
                           polishing_kwargs_reflectivity: dict = None,
                           ) -> dict:
        params = predicted_params.parameters.squeeze().cpu().numpy()

        polished_params_dict = {}
        polishing_kwargs_reflectivity = polishing_kwargs_reflectivity or {}

        try:
            if fit_growth:
                polished_params_arr, curve_polished = get_fit_with_growth(
                    q = q, 
                    curve = curve, 
                    init_params = params, 
                    bounds = priors.T,
                    max_d_change = max_d_change,
                )
                polished_params = BasicParams(
                    torch.from_numpy(polished_params_arr[:-1][None]),
                    torch.from_numpy(priors.T[0][None]),
                    torch.from_numpy(priors.T[1][None]),
                    self.trainer.loader.prior_sampler.max_num_layers,
                    self.trainer.loader.prior_sampler.param_model
                    )
            else:
                polished_params_arr, curve_polished = refl_fit(
                    q = q, 
                    curve = curve, 
                    init_params = params, 
                    bounds=priors.T,
                    prior_sampler=self.trainer.loader.prior_sampler,
                    error_bars=error_bars,
                    method=polishing_method,
                    polishing_max_nfev=polishing_max_nfev,
                    reflectivity_kwargs=polishing_kwargs_reflectivity,
                )
                polished_params = BasicParams(
                    torch.from_numpy(polished_params_arr[None]),
                    torch.from_numpy(priors.T[0][None]),
                    torch.from_numpy(priors.T[1][None]),
                    self.trainer.loader.prior_sampler.max_num_layers,
                    self.trainer.loader.prior_sampler.param_model
                )
        except Exception as err:
            polished_params = predicted_params
            polished_params_arr = get_prediction_array(polished_params)
            curve_polished = np.zeros_like(q)

        polished_params_dict['polished_params_array'] = polished_params_arr
        if calc_polished_curve:
            polished_params_dict['polished_curve'] = curve_polished

        if ambient_sld_tensor is not None:
            ambient_sld_tensor = ambient_sld_tensor.to(polished_params.slds.device)


        if calc_polished_sld_profile:
            _, sld_profile_polished, _ = get_density_profiles(
                polished_params.thicknesses, polished_params.roughnesses, polished_params.slds + (ambient_sld_tensor or 0), ambient_sld_tensor, 
                z_axis=sld_x_axis.to(polished_params.slds.device),
            )
            polished_params_dict['sld_profile_polished'] = sld_profile_polished.squeeze().cpu().numpy()

        return polished_params_dict
    
    def _scale_curve(self, curve: Union[np.ndarray, Tensor]):
        if not isinstance(curve, Tensor):
            curve = torch.from_numpy(curve).float()
        curve = curve.unsqueeze(0).to(self.device)
        scaled_curve = self.trainer.loader.curves_scaler.scale(curve)
        return scaled_curve
    
    def _scale_prior_bounds(self, prior_bounds: List[Tuple]):
        try:
            prior_bounds = torch.tensor(prior_bounds)
            prior_bounds = prior_bounds.to(self.device).T
            min_bounds, max_bounds = prior_bounds[:, None]

            scaled_bounds = torch.cat([
                self.trainer.loader.prior_sampler.scale_bounds(min_bounds), 
                self.trainer.loader.prior_sampler.scale_bounds(max_bounds)
            ], -1)

            return scaled_bounds.float()

        except RuntimeError as e:
            expected_param_dim = self.trainer.loader.prior_sampler.param_dim
            actual_param_dim = prior_bounds.shape[1] if prior_bounds.ndim == 2 else len(prior_bounds)

            msg = (
                f"\n **Parameter dimension mismatch during inference!**\n"
                f"- Model expects **{expected_param_dim}** parameters.\n"
                f"- You provided **{actual_param_dim}** prior bounds.\n\n"
                f"💡This often occurs when:\n"
                f"- The model was trained with additional nuisance parameters like `r_scale`, `q_shift`, or `log10_background`,\n"
                f"  but they were not included in the `prior_bounds` passed to `.predict()`.\n"
                f"- The number of layers or parameterization type differs from the one used during training.\n\n"
                f" Check the configuration or the summary of expected parameters."
            )
            raise ValueError(msg) from e    
    
    def _shift_slds_by_ambient(self, prior_bounds: np.ndarray, ambient_sld: float):
        n_layers = self.trainer.loader.prior_sampler.max_num_layers
        sld_indices = slice(2*n_layers+1, 3*n_layers+2)
        prior_bounds[sld_indices, ...] -= ambient_sld

        training_min_bounds = self.trainer.loader.prior_sampler.min_bounds.squeeze().cpu().numpy()
        training_max_bounds = self.trainer.loader.prior_sampler.max_bounds.squeeze().cpu().numpy()
        lower_bound_check = (prior_bounds[sld_indices, 0] >= training_min_bounds[sld_indices]).all()
        upper_bound_check = (prior_bounds[sld_indices, 1] <= training_max_bounds[sld_indices]).all()
        assert lower_bound_check and upper_bound_check, "Shifting the layer SLDs by the ambient SLD exceeded the training ranges."

        return sld_indices
    
    def _restore_slds_after_ambient_shift(self, prediction_dict, sld_indices, ambient_sld):
        prediction_dict["predicted_params_array"][sld_indices] += ambient_sld
        if "polished_params_array" in prediction_dict:
            prediction_dict["polished_params_array"][sld_indices] += ambient_sld

    def _get_likelihood(self, q, curve, rel_err: float = 0.1, abs_err: float = 1e-12):
        return LogLikelihood(
            q, curve, self.trainer.loader.prior_sampler, curve * rel_err + abs_err
        )
    
    def get_param_labels(self, **kwargs):
        return self.trainer.loader.prior_sampler.param_model.get_param_labels(**kwargs)
    
    @staticmethod
    def _preprocess_input_data(
                           reflectivity_curve,
                           q_values,
                           sigmas=None,
                           q_resolution=None,
                           truncate_index_left=None,
                           truncate_index_right=None,
                           enable_error_bars_filtering=True,
                           filter_threshold=0.3,
                           filter_remove_singles=True,
                           filter_remove_consecutives=True,
                           filter_consecutive=3,
                           filter_q_start_trunc=0.1):
    
        # Save originals for polishing
        reflectivity_curve_original = reflectivity_curve.copy()
        q_values_original = q_values.copy() if q_values is not None else None
        q_resolution_original = q_resolution.copy() if isinstance(q_resolution, np.ndarray) else q_resolution
        sigmas_original = sigmas.copy() if sigmas is not None else None

        # Remove points with non-positive intensities
        nonnegative_mask = reflectivity_curve > 0.0
        reflectivity_curve = reflectivity_curve[nonnegative_mask]
        q_values = q_values[nonnegative_mask]
        if sigmas is not None:
            sigmas = sigmas[nonnegative_mask]
        if isinstance(q_resolution, np.ndarray):
            q_resolution = q_resolution[nonnegative_mask]

        # Truncate arrays
        if truncate_index_left is not None or truncate_index_right is not None:
            slice_obj = slice(truncate_index_left, truncate_index_right)
            reflectivity_curve = reflectivity_curve[slice_obj]
            q_values = q_values[slice_obj]
            if sigmas is not None:
                sigmas = sigmas[slice_obj]
            if isinstance(q_resolution, np.ndarray):
                q_resolution = q_resolution[slice_obj]

        # Filter high-error points
        if enable_error_bars_filtering and sigmas is not None:
            valid_mask = get_filtering_mask(
                q_values,
                reflectivity_curve,
                sigmas,
                threshold=filter_threshold,
                consecutive=filter_consecutive,
                remove_singles=filter_remove_singles,
                remove_consecutives=filter_remove_consecutives,
                q_start_trunc=filter_q_start_trunc
            )
            reflectivity_curve = reflectivity_curve[valid_mask]
            q_values = q_values[valid_mask]
            sigmas = sigmas[valid_mask]
            if isinstance(q_resolution, np.ndarray):
                q_resolution = q_resolution[valid_mask]

        return (q_values, reflectivity_curve, sigmas, q_resolution,
                q_values_original, reflectivity_curve_original,
                sigmas_original, q_resolution_original)
    
    def interpolate_data_to_model_q(
        self,
        q_exp,
        refl_exp,
        sigmas_exp=None,
        q_res_exp=None,
        as_dict=False
    ):
        q_generator = self.trainer.loader.q_generator

        def _pad(arr, pad_to, value=0.0):
            if arr is None:
                return None
            return np.pad(arr, (0, pad_to - len(arr)), constant_values=value)

        def _interp_or_keep(q_model, q_exp, arr):
            """Interpolate arrays, keep floats or None unchanged."""
            if arr is None:
                return None
            return np.interp(q_model, q_exp, arr) if isinstance(arr, np.ndarray) else arr

        def _pad_or_keep(arr, max_n):
            """Pad arrays, keep floats or None unchanged."""
            if arr is None:
                return None
            return _pad(arr, max_n, 0.0) if isinstance(arr, np.ndarray) else arr

        def _prepare_return(q, refl, sigmas=None, q_res=None, mask=None, as_dict=False):
            if as_dict:
                result = {"q_model": q, "reflectivity": refl}
                if sigmas is not None: result["sigmas"] = sigmas
                if q_res is not None: result["q_resolution"] = q_res
                if mask is not None: result["key_padding_mask"] = mask
                return result
            result = [q, refl]
            if sigmas is not None: result.append(sigmas)
            if q_res is not None: result.append(q_res)
            if mask is not None: result.append(mask)
            return tuple(result)

        # ConstantQ
        if isinstance(q_generator, ConstantQ):
            q_model = q_generator.q.cpu().numpy()
            refl_out = interp_reflectivity(q_model, q_exp, refl_exp)
            sigmas_out = _interp_or_keep(q_model, q_exp, sigmas_exp)
            q_res_out = _interp_or_keep(q_model, q_exp, q_res_exp)
            return _prepare_return(q_model, refl_out, sigmas_out, q_res_out, None, as_dict)

        # VariableQ
        elif isinstance(q_generator, VariableQ):
            if q_generator.n_q_range[0] == q_generator.n_q_range[1]:
                n_q_model = q_generator.n_q_range[0]
                q_min = max(q_exp.min(), q_generator.q_min_range[0])
                q_max = min(q_exp.max(), q_generator.q_max_range[1])
                if self.trainer.loader.q_generator.mode == 'logspace':
                    q_model = torch.logspace(start=torch.log10(torch.tensor(q_min, device=self.device)),
                                             end=torch.log10(torch.tensor(q_max, device=self.device)),
                                             steps=n_q_model, device=self.device).to('cpu')
                    logspace = True
                else:
                    q_model = np.linspace(q_min, q_max, n_q_model)
                    logspace = False
            else:
                return _prepare_return(q_exp, refl_exp, sigmas_exp, q_res_exp, None, as_dict)

            refl_out = interp_reflectivity(q_model, q_exp, refl_exp, logspace=logspace)
            sigmas_out = _interp_or_keep(q_model, q_exp, sigmas_exp)
            q_res_out = _interp_or_keep(q_model, q_exp, q_res_exp)
            return _prepare_return(q_model, refl_out, sigmas_out, q_res_out, None, as_dict)

        # MaskedVariableQ
        elif isinstance(q_generator, MaskedVariableQ):
            min_n, max_n = q_generator.n_q_range
            n_exp = len(q_exp)

            if min_n <= n_exp <= max_n:
                # Pad only
                q_model = _pad(q_exp, max_n, 0.0)
                refl_out = _pad(refl_exp, max_n, 0.0)
                sigmas_out = _pad_or_keep(sigmas_exp, max_n)
                q_res_out = _pad_or_keep(q_res_exp, max_n)
                key_padding_mask = np.zeros(max_n, dtype=bool)
                key_padding_mask[:n_exp] = True

            else:
                # Interpolate + pad
                n_interp = min(max(n_exp, min_n), max_n)
                q_min = max(q_exp.min(), q_generator.q_min_range[0])
                q_max = min(q_exp.max(), q_generator.q_max_range[1])
                q_interp = np.linspace(q_min, q_max, n_interp)

                refl_interp = interp_reflectivity(q_interp, q_exp, refl_exp)
                sigmas_interp = _interp_or_keep(q_interp, q_exp, sigmas_exp)
                q_res_interp = _interp_or_keep(q_interp, q_exp, q_res_exp)

                q_model = _pad(q_interp, max_n, 0.0)
                refl_out = _pad(refl_interp, max_n, 0.0)
                sigmas_out = _pad_or_keep(sigmas_interp, max_n)
                q_res_out = _pad_or_keep(q_res_interp, max_n)
                key_padding_mask = np.zeros(max_n, dtype=bool)
                key_padding_mask[:n_interp] = True

            return _prepare_return(q_model, refl_out, sigmas_out, q_res_out, key_padding_mask, as_dict)

        else:
            raise TypeError(f"Unsupported QGenerator type: {type(q_generator)}")
        
    def _qshift_prediction(self, curve, scaled_bounds, num: int = 1000, dq_coef: float = 1.) -> BasicParams:
        assert isinstance(self.trainer.loader.q_generator, ConstantQ), "Prediction with q shifts available only for models with fixed discretization"
        q = self.trainer.loader.q_generator.q.squeeze().float()
        dq_max = (q[1] - q[0]) * dq_coef
        q_shifts = torch.linspace(-dq_max, dq_max, num).to(q)

        curve = to_t(curve).to(scaled_bounds)
        shifted_curves = _qshift_interp(q.squeeze(), curve, q_shifts)

        assert shifted_curves.shape == (num, q.shape[0])

        scaled_curves = self.trainer.loader.curves_scaler.scale(shifted_curves)
        scaled_prior_bounds = torch.atleast_2d(scaled_bounds).expand(scaled_curves.shape[0], -1)

        with torch.no_grad():
            self.trainer.model.eval()
            scaled_predicted_params = self.trainer.model(scaled_curves, scaled_prior_bounds)
            restored_params = self.trainer.loader.prior_sampler.restore_params(torch.cat([scaled_predicted_params, scaled_prior_bounds], dim=-1))

            best_param = get_best_mse_param(
                restored_params,
                self._get_likelihood(q=self.trainer.loader.q_generator.q, curve=curve),
            )
            return best_param

    def predict_using_widget(self, reflectivity_curve, **kwargs):
        """
        """

        NUM_INTERVALS = self.trainer.loader.prior_sampler.param_dim
        param_labels = self.trainer.loader.prior_sampler.param_model.get_param_labels()
        min_bounds = self.trainer.loader.prior_sampler.min_bounds.cpu().numpy().flatten()
        max_bounds = self.trainer.loader.prior_sampler.max_bounds.cpu().numpy().flatten()
        max_deltas = self.trainer.loader.prior_sampler.max_delta.cpu().numpy().flatten()

        print(f'Adjust the sliders for each parameter and press "Predict". Repeat as desired. Press "Close Widget" to finish.')

        interval_widgets = []
        for i in range(NUM_INTERVALS):
            label = widgets.Label(value=f'{param_labels[i]}')
            initial_max = min(max_bounds[i], min_bounds[i] + max_deltas[i])
            slider = widgets.FloatRangeSlider(
                value=[min_bounds[i], initial_max],
                min=min_bounds[i],
                max=max_bounds[i],
                step=0.01,
                layout=widgets.Layout(width='400px'),
                style={'description_width': '60px'}
            )

            def validate_range(change, slider=slider, max_width=max_deltas[i]):
                min_val, max_val = change['new']
                if max_val - min_val > max_width:
                    old_min_val, old_max_val = change['old']
                    if abs(old_min_val - min_val) > abs(old_max_val - max_val):
                        max_val = min_val + max_width
                    else:
                        min_val = max_val - max_width
                    slider.value = [min_val, max_val]

            slider.observe(validate_range, names='value')
            interval_widgets.append((slider, widgets.HBox([label, slider])))

        sliders_box = widgets.VBox([iw[1] for iw in interval_widgets])

        output = widgets.Output()
        predict_button = widgets.Button(description="Predict")
        close_button = widgets.Button(description="Close Widget")

        container = widgets.VBox([sliders_box, widgets.HBox([predict_button, close_button]), output])
        display(container)

        @output.capture(clear_output=True)
        def on_predict_click(_):
            if 'prior_bounds' in kwargs:
                array_values = kwargs.pop('prior_bounds')
                for i, (s, _) in enumerate(interval_widgets):
                    s.value = tuple(array_values[i])
            else:
                values = [(s.value[0], s.value[1]) for s, _ in interval_widgets]
                array_values = np.array(values)

            prediction_result = self.predict(reflectivity_curve=reflectivity_curve,
                                            prior_bounds=array_values,
                                            **kwargs)
            param_names = self.trainer.loader.prior_sampler.param_model.get_param_labels()
            print_prediction_results(prediction_result)

            plot_prediction_results(
                prediction_result,
                q_exp=kwargs['q_values'],
                curve_exp=reflectivity_curve,
            )
            self.widget_prediction_result = prediction_result

        def on_close_click(_):
            container.close()
            print("Widget closed.")

        predict_button.on_click(on_predict_click)
        close_button.on_click(on_close_click)

    def preprocess_and_predict_using_widget(self,
                                            reflectivity_curve,
                                            q_values=None,
                                            sigmas=None,
                                            q_resolution=None,
                                            prior_bounds=None,
                                            ambient_sld=None,
                                            ):
        """
        Interactive widget around `preprocess_and_predict`
        Results are stored in `self.widget_prediction_result`
        """

        if q_values is None:
            raise ValueError("q_values must be provided for this widget.")

        N = len(reflectivity_curve)

        # ---------- Priors sliders ----------
        param_labels = self.trainer.loader.prior_sampler.param_model.get_param_labels()
        min_bounds = self.trainer.loader.prior_sampler.min_bounds.cpu().numpy().flatten()
        max_bounds = self.trainer.loader.prior_sampler.max_bounds.cpu().numpy().flatten()
        max_deltas = self.trainer.loader.prior_sampler.max_delta.cpu().numpy().flatten()
        NUM = len(param_labels)

        sliders, rows = [], []
        init_pb = np.array(prior_bounds) if prior_bounds is not None else None
        for i in range(NUM):
            init_min = float(init_pb[i, 0]) if init_pb is not None else float(min_bounds[i])
            init_max = float(init_pb[i, 1]) if init_pb is not None else float(min(min_bounds[i] + max_deltas[i], max_bounds[i]))
            lab = widgets.Label(value=param_labels[i])
            s = widgets.FloatRangeSlider(
                value=[init_min, init_max],
                min=float(min_bounds[i]),
                max=float(max_bounds[i]),
                step=0.01,
                layout=widgets.Layout(width='420px'),
                readout_format='.3f'
            )
            # Constrain slider widths
            def _mk_validator(slider, max_width=float(max_deltas[i])):
                def _validate(change):
                    a, b = change['new']
                    if b - a > max_width:
                        oa, ob = change['old']
                        if abs(oa - a) > abs(ob - b):
                            b = a + max_width
                        else:
                            a = b - max_width
                        slider.value = (a, b)
                return _validate
            s.observe(_mk_validator(s), names='value')
            sliders.append(s)
            rows.append(widgets.HBox([lab, s]))
        priors_box = widgets.VBox([widgets.HTML("<b>Priors</b>")] + rows)

        # ---------- Preprocess & Predict controls ----------
        # Preprocessing
        trunc_L = widgets.IntSlider(description='truncate left',  min=0, max=max(0, N-1), step=1, value=0)
        trunc_R = widgets.IntSlider(description='truncate right', min=1, max=N,           step=1, value=N)
        enable_filt = widgets.Checkbox(description='filter error bars', value=True)
        thr = widgets.FloatSlider(description='filtering threshold', min=0.0, max=1.0, step=0.01, value=0.3)
        rem_single = widgets.Checkbox(description='remove singles', value=True)
        rem_cons = widgets.Checkbox(description='remove consecutives', value=True)
        consec = widgets.IntSlider(description='num. consecutive', min=1, max=10, step=1, value=3)
        qstart = widgets.FloatSlider(description='q_start_trunc', min=0.0, max=1.0, step=0.01, value=0.1)

        # Polishing
        polish = widgets.Checkbox(description='polish prediction', value=True)
        use_sigmas_polish = widgets.Checkbox(description='use sigmas during polishing', value=True)

        # Plotting
        pred_show_yerr = widgets.Checkbox(description='show error bars', value=True)
        pred_show_xerr = widgets.Checkbox(description='show q-resolution', value=False)
        pred_logx = widgets.Checkbox(description='log x-axis', value=False)
        plot_sld = widgets.Checkbox(description='plot SLD profile', value=True)
        sld_pad_left = widgets.FloatText(description='SLD pad left',  value=0.2, step=0.1)
        sld_pad_right = widgets.FloatText(description='SLD pad right', value=1.1, step=0.1)

        # Color pickers
        exp_color_picker = widgets.ColorPicker(description='exp color',       value='#0000FF')  # blue
        exp_errcolor_picker = widgets.ColorPicker(description='errbar color',    value='#800080')  # purple
        pred_color_picker = widgets.ColorPicker(description='pred color',      value='#FF0000')  # red
        pol_color_picker = widgets.ColorPicker(description='polished color',  value='#FFA500')  # orange
        sld_pred_color_picker = widgets.ColorPicker(description='SLD pred color',  value='#FF0000')  # red
        sld_pol_color_picker = widgets.ColorPicker(description='SLD pol color',   value='#FFA500')  # orange

        # Compute toggles
        calc_curve = widgets.Checkbox(description='calc curve', value=True)
        calc_sld_pred = widgets.Checkbox(description='calc predicted SLD', value=True)
        calc_sld_pol = widgets.Checkbox(description='calc polished SLD', value=True)

        btn_predict = widgets.Button(description='Predict')
        btn_close = widgets.Button(description='Close')

        controls_box = widgets.VBox([
            widgets.HTML("<b>Preprocess & Predict</b>"),
            widgets.HTML("<i>Preprocessing</i>"),
            widgets.HBox([trunc_L, trunc_R]),
            widgets.HBox([enable_filt, rem_single, rem_cons]),
            widgets.HBox([thr, consec, qstart]),
            widgets.HTML("<i>Polishing</i>"),
            widgets.HBox([polish, use_sigmas_polish]),
            widgets.HTML("<i>Plotting</i>"),
            widgets.HBox([pred_show_yerr, pred_show_xerr, pred_logx, plot_sld]),
            widgets.HBox([sld_pad_left, sld_pad_right]),
            widgets.HTML("<i>Colors</i>"),
            widgets.HBox([exp_color_picker, exp_errcolor_picker]),
            widgets.HBox([pred_color_picker, pol_color_picker]),
            widgets.HBox([sld_pred_color_picker, sld_pol_color_picker]),
            widgets.HTML("<i>Compute</i>"),
            widgets.HBox([calc_curve, calc_sld_pred, calc_sld_pol]),
            widgets.HBox([btn_predict, btn_close]),
        ])

        out_predict = widgets.Output()
        container = widgets.VBox([priors_box, controls_box, out_predict])
        display(container)

        def _sync_trunc(_):
            if trunc_L.value >= trunc_R.value:
                trunc_L.value = max(0, trunc_R.value - 1)
        trunc_L.observe(_sync_trunc, names='value')
        trunc_R.observe(_sync_trunc, names='value')

        def _current_priors():
            return np.array([s.value for s in sliders], dtype=np.float32)  # (param_dim, 2)

        @out_predict.capture(clear_output=True)
        def _on_predict(_):
            out_predict.clear_output(wait=True)

            res = self.preprocess_and_predict(
                reflectivity_curve=reflectivity_curve,
                q_values=q_values,
                prior_bounds=_current_priors(),
                sigmas=sigmas,
                q_resolution=q_resolution,
                ambient_sld=ambient_sld,
                clip_prediction=True,
                polish_prediction=polish.value,
                use_sigmas_for_polishing=use_sigmas_polish.value,
                calc_pred_curve=calc_curve.value,
                calc_pred_sld_profile=(calc_sld_pred.value or plot_sld.value),
                calc_polished_sld_profile=(calc_sld_pol.value or plot_sld.value),
                sld_profile_padding_left=float(sld_pad_left.value),
                sld_profile_padding_right=float(sld_pad_right.value),

                truncate_index_left=trunc_L.value,
                truncate_index_right=trunc_R.value,
                enable_error_bars_filtering=enable_filt.value,
                filter_threshold=thr.value,
                filter_remove_singles=rem_single.value,
                filter_remove_consecutives=rem_cons.value,
                filter_consecutive=consec.value,
                filter_q_start_trunc=qstart.value,
            )

            # Full experimental data as scatter
            q_exp_plot = q_values
            r_exp_plot = reflectivity_curve
            yerr_plot = (sigmas if pred_show_yerr.value else None)
            xerr_plot = (q_resolution if pred_show_xerr.value else None)

            # Predicted curve only on the model region
            q_pred = res.get('q_plot_pred', None)
            r_pred = res.get('predicted_curve', None)

            # Polished curve on the full experimental grid
            q_pol = q_values if ('polished_curve' in res) else None
            r_pol = res.get('polished_curve', None)

            # SLD profiles
            z_sld = res.get('predicted_sld_xaxis', None)
            sld_pred = res.get('predicted_sld_profile', None)
            sld_pol = res.get('sld_profile_polished', None)

            print_prediction_results(res)

            plot_reflectivity(
                q_exp=q_exp_plot, r_exp=r_exp_plot,
                yerr=yerr_plot, xerr=xerr_plot,
                exp_style=('errorbar' if pred_show_yerr.value or pred_show_xerr.value else 'scatter'),
                exp_color=exp_color_picker.value,
                exp_errcolor=exp_errcolor_picker.value,
                q_pred=q_pred, r_pred=r_pred, pred_color=pred_color_picker.value,
                q_pol=q_pol, r_pol=r_pol, pol_color=pol_color_picker.value,
                z_sld=z_sld, sld_pred=sld_pred, sld_pol=sld_pol,
                sld_pred_color=sld_pred_color_picker.value,
                sld_pol_color=sld_pol_color_picker.value,
                plot_sld_profile=plot_sld.value,
                logx=pred_logx.value, logy=True,
                figsize=(12,6),
                legend=True
            )

            self.widget_prediction_result = res

        def _on_close(_):
            container.close()
            print("Widget closed.")

        btn_predict.on_click(_on_predict)
        btn_close.on_click(_on_close)

class InferenceModel(object):
    def __init__(self, name: str = None, trainer: PointEstimatorTrainer = None, preprocessing_parameters: dict = None,
                 num_sampling: int = 2 ** 13):
        self.log = logging.getLogger(__name__)
        self.model_name = name
        self.trainer = trainer
        self.q = None
        self.preprocessing = StandardPreprocessing(**(preprocessing_parameters or {}))
        self._sampling_num = num_sampling

        if trainer is None and self.model_name is not None:
            self.load_model(self.model_name)
        elif trainer is not None:
            self._set_trainer(trainer, preprocessing_parameters)

    ### API methods ###

    def load_model(self, name: str) -> None:
        self.log.debug(f"loading model {name}")
        if self.model_name == name and self.trainer is not None:
            return
        self.model_name = name
        self._set_trainer(get_trainer_by_name(name))
        self.log.info(f"Model {name} is loaded.")

    def train_model(self, name: str):
        self.model_name = name
        self.trainer = train_from_config(load_config(name))

    def set_preprocessing_parameters(self, **kwargs) -> None:
        self.preprocessing.set_parameters(**kwargs)

    def preprocess(self,
                   intensity: np.ndarray,
                   scattering_angle: np.ndarray,
                   attenuation: np.ndarray,
                   update_params: bool = False,
                   **kwargs) -> dict:
        if update_params:
            self.preprocessing.set_parameters(**kwargs)
        preprocessed_dict = self.preprocessing(intensity, scattering_angle, attenuation, **kwargs)
        return preprocessed_dict

    def predict(self,
                intensity: np.ndarray,
                scattering_angle: np.ndarray,
                attenuation: np.ndarray,
                priors: np.ndarray,
                preprocessing_parameters: dict = None,
                polish: bool = True,
                use_sampler: bool = False,
                use_q_shift: bool = True,
                max_d_change: float = 5.,
                fit_growth: bool = True,
                ) -> dict:

        with print_time("everything"):
            with print_time("preprocess"):
                preprocessed_dict = self.preprocess(
                    intensity, scattering_angle, attenuation, **(preprocessing_parameters or {})
                )

            preprocessed_curve = preprocessed_dict["curve_interp"]
            raw_curve, raw_q = preprocessed_dict["curve"], preprocessed_dict["q_values"]
            q_ratio = preprocessed_dict["q_ratio"]

            with print_time("predict_from_preprocessed_curve"):
                preprocessed_dict.update(self.predict_from_preprocessed_curve(
                    preprocessed_curve, priors, raw_curve=raw_curve, raw_q=raw_q, polish=polish, q_ratio=q_ratio,
                    use_sampler=use_sampler, use_q_shift=use_q_shift, max_d_change=max_d_change,
                    fit_growth=fit_growth,
                ))

            return preprocessed_dict

    def predict_from_preprocessed_curve(self,
                                        curve: np.ndarray,
                                        priors: np.ndarray, *,
                                        polish: bool = True,
                                        raw_curve: np.ndarray = None,
                                        raw_q: np.ndarray = None,
                                        clip_prediction: bool = True,
                                        q_ratio: float = 1.,
                                        use_sampler: bool = False,
                                        use_q_shift: bool = True,
                                        max_d_change: float = 5.,
                                        fit_growth: bool = True,
                                        ) -> dict:

        scaled_curve = self._scale_curve(curve)
        scaled_bounds, min_bounds, max_bounds = self._scale_priors(priors, q_ratio)

        if not use_q_shift:
            predicted_params: UniformSubPriorParams = self._simple_prediction(scaled_curve, scaled_bounds)
        else:
            predicted_params: UniformSubPriorParams = self._qshift_prediction(curve, scaled_bounds)

        if use_sampler:
            predicted_params: UniformSubPriorParams = self._sampler_solution(
                curve, predicted_params,
            )

        if clip_prediction:
            predicted_params = self._prior_sampler.clamp_params(predicted_params)

        if raw_curve is None:
            raw_curve = curve
        if raw_q is None:
            raw_q = self.q.squeeze().cpu().numpy()
            raw_q_t = self.q
        else:
            raw_q_t = torch.from_numpy(raw_q).to(self.q)

        if q_ratio != 1.:
            predicted_params.scale_with_q(q_ratio)
            raw_q = raw_q * q_ratio
            raw_q_t = raw_q_t * q_ratio

        prediction_dict = {
            "params": get_prediction_array(predicted_params),
            "param_names": get_param_labels(
                predicted_params.max_layer_num,
                thickness_name='d',
                roughness_name='sigma',
                sld_name='rho',
            ),
            "curve_predicted": predicted_params.reflectivity(raw_q_t).squeeze().cpu().numpy()
        }

        sld_x_axis, sld_profile, _ = get_density_profiles(
            predicted_params.thicknesses, predicted_params.roughnesses, predicted_params.slds, num=1024,
        )

        prediction_dict['sld_profile'] = sld_profile.squeeze().cpu().numpy()
        prediction_dict['sld_x_axis'] = sld_x_axis.squeeze().cpu().numpy()

        if polish:
            prediction_dict.update(self._polish_prediction(
                raw_q, raw_curve, predicted_params, priors, sld_x_axis,
                max_d_change=max_d_change, fit_growth=fit_growth,
            ))

            if fit_growth and "params_polished" in prediction_dict:
                prediction_dict["param_names"].append("max_d_change")

        return prediction_dict

    ### some shortcut methods for data processing ###

    def _simple_prediction(self, scaled_curve, scaled_bounds) -> UniformSubPriorParams:
        context = torch.cat([scaled_curve, scaled_bounds], -1)

        with torch.no_grad():
            self.trainer.model.eval()
            scaled_params = self.trainer.model(context)

        predicted_params: UniformSubPriorParams = self._restore_predicted_params(scaled_params, context)
        return predicted_params

    @print_time
    def _qshift_prediction(self, curve, scaled_bounds, num: int = 1000, dq_coef: float = 1.) -> UniformSubPriorParams:
        q = self.q.squeeze().float()
        curve = to_t(curve).to(q)
        dq_max = (q[1] - q[0]) * dq_coef
        q_shifts = torch.linspace(-dq_max, dq_max, num).to(q)
        shifted_curves = _qshift_interp(q.squeeze(), curve, q_shifts)

        assert shifted_curves.shape == (num, q.shape[0])

        scaled_curves = self.trainer.loader.curves_scaler.scale(shifted_curves)
        context = torch.cat([scaled_curves, torch.atleast_2d(scaled_bounds).expand(scaled_curves.shape[0], -1)], -1)

        with torch.no_grad():
            self.trainer.model.eval()
            scaled_params = self.trainer.model(context)
            restored_params = self._restore_predicted_params(scaled_params, context)

            best_param = get_best_mse_param(
                restored_params,
                self._get_likelihood(curve),
            )
            return best_param

    @print_time
    def _polish_prediction(self,
                           q: np.ndarray,
                           curve: np.ndarray,
                           predicted_params: Params,
                           priors: np.ndarray,
                           sld_x_axis,
                           fit_growth: bool = True,
                           max_d_change: float = 5.,
                           ) -> dict:
        params = torch.cat([
            predicted_params.thicknesses.squeeze(),
            predicted_params.roughnesses.squeeze(),
            predicted_params.slds.squeeze()
        ]).cpu().numpy()

        polished_params_dict = {}

        try:
            if fit_growth:
                polished_params_arr, curve_polished = get_fit_with_growth(
                    q, curve, params, bounds=priors.T,
                    max_d_change=max_d_change,
                )
                polished_params = Params.from_tensor(torch.from_numpy(polished_params_arr[:-1][None]).to(self.q))
            else:
                polished_params_arr, curve_polished = standard_refl_fit(q, curve, params, bounds=priors.T)
                polished_params = Params.from_tensor(torch.from_numpy(polished_params_arr[None]).to(self.q))
        except Exception as err:
            self.log.exception(err)
            polished_params = predicted_params
            polished_params_arr = get_prediction_array(polished_params)
            curve_polished = np.zeros_like(q)

        polished_params_dict['params_polished'] = polished_params_arr
        polished_params_dict['curve_polished'] = curve_polished

        sld_x_axis_polished, sld_profile_polished, _ = get_density_profiles(
            polished_params.thicknesses, polished_params.roughnesses, polished_params.slds, z_axis=sld_x_axis,
        )

        polished_params_dict['sld_profile_polished'] = sld_profile_polished.squeeze().cpu().numpy()

        return polished_params_dict

    def _restore_predicted_params(self, scaled_params: Tensor, context: Tensor) -> UniformSubPriorParams:
        predicted_params: UniformSubPriorParams = self.trainer.loader.prior_sampler.restore_params(
            self.trainer.loader.prior_sampler.PARAM_CLS.restore_params_from_context(scaled_params, context)
        )
        return predicted_params

    def _input2context(self, curve: np.ndarray, priors: np.ndarray, q_ratio: float = 1.):
        scaled_curve = self._scale_curve(curve)
        scaled_bounds, min_bounds, max_bounds = self._scale_priors(priors, q_ratio)
        scaled_input = torch.cat([scaled_curve, scaled_bounds], -1)
        return scaled_input, min_bounds, max_bounds

    def _scale_curve(self, curve: np.ndarray or Tensor):
        if not isinstance(curve, Tensor):
            curve = torch.from_numpy(curve).float()
        curve = torch.atleast_2d(curve).to(self.q)
        scaled_curve = self.trainer.loader.curves_scaler.scale(curve)
        return scaled_curve.float()

    def _scale_priors(self, priors: np.ndarray or Tensor, q_ratio: float = 1.):
        if not isinstance(priors, Tensor):
            priors = torch.from_numpy(priors)

        priors = priors.float().clone()

        priors = priors.to(self.q).T
        priors = self._prior_sampler.scale_bounds_with_q(priors, 1 / q_ratio)
        priors = self._prior_sampler.clamp_bounds(priors)

        min_bounds, max_bounds = priors[:, None].to(self.q)
        prior_sampler = self._prior_sampler
        scaled_bounds = torch.cat([
            prior_sampler.scale_bounds(min_bounds), prior_sampler.scale_bounds(max_bounds)
        ], -1)
        return scaled_bounds.float(), min_bounds, max_bounds

    @property
    def _prior_sampler(self) -> ExpUniformSubPriorSampler:
        return self.trainer.loader.prior_sampler

    def _set_trainer(self, trainer, preprocessing_parameters: dict = None):
        self.trainer = trainer
        self.trainer.model.eval()
        self._update_preprocessing(preprocessing_parameters)

    def _update_preprocessing(self, preprocessing_parameters: dict = None):
        self.log.debug(f"setting preprocessing_parameters {preprocessing_parameters}.")
        self.q = self.trainer.loader.q_generator.q
        self.preprocessing = StandardPreprocessing(
            self.q.cpu().squeeze().numpy(),
            **(preprocessing_parameters or {})
        )
        self.log.info(f"preprocessing params are set: {preprocessing_parameters}.")

    @print_time
    def _sampler_solution(
            self,
            curve: Tensor or np.ndarray,
            predicted_params: UniformSubPriorParams,
    ) -> UniformSubPriorParams:

        if not isinstance(curve, Tensor):
            curve = torch.from_numpy(curve).float()
        curve = curve.to(self.q)

        refined_params = simple_sampler_solution(
            self._get_likelihood(curve),
            predicted_params,
            self._prior_sampler.min_bounds,
            self._prior_sampler.max_bounds,
            num=self._sampling_num, coef=0.1,
        )

        return refined_params

    def _get_likelihood(self, curve, rel_err: float = 0.1, abs_err: float = 1e-12):
        return LogLikelihood(
            self.q, curve, self._prior_sampler, curve * rel_err + abs_err
        )


def get_prediction_array(params: BasicParams) -> np.ndarray:
    predict_arr = torch.cat([
        params.thicknesses.squeeze(),
        params.roughnesses.squeeze(),
        params.slds.squeeze(),
    ]).cpu().numpy()

    return predict_arr


def _qshift_interp(q, r, q_shifts):
    qs = q[None] + q_shifts[:, None]
    eps = torch.finfo(r.dtype).eps
    ind = torch.searchsorted(q[None].expand_as(qs).contiguous(), qs.contiguous())
    ind = torch.clamp(ind - 1, 0, q.shape[0] - 2)
    slopes = (r[1:] - r[:-1]) / (eps + (q[1:] - q[:-1]))
    return r[ind] + slopes[ind] * (qs - q[ind])


from torch import Tensor

from reflectorch.data_generation.priors.params import Params

__all__ = [
    "PriorSampler",
]


class PriorSampler(object):
    """Base class for prior samplers"""
    
    PARAM_CLS = Params

    @property
    def param_dim(self) -> int:
        """gets the number of parameters (i.e. the parameter dimensionality)"""
        return self.PARAM_CLS.layers_num2size(self.max_num_layers)

    @property
    def max_num_layers(self) -> int:
        """gets the number of layers"""
        raise NotImplementedError

    def sample(self, batch_size: int) -> Params:
        """sample a batch of parameters"""
        raise NotImplementedError

    def scale_params(self, params: Params) -> Tensor:
        """scale the parameters to a ML-friendly range"""
        raise NotImplementedError

    def restore_params(self, scaled_params: Tensor) -> Params:
        """restore the parameters to their original range"""
        raise NotImplementedError

    def log_prob(self, params: Params) -> Tensor:
        raise NotImplementedError

    def get_indices_within_domain(self, params: Params) -> Tensor:
        raise NotImplementedError

    def get_indices_within_bounds(self, params: Params) -> Tensor:
        raise NotImplementedError

    def filter_params(self, params: Params) -> Params:
        indices = self.get_indices_within_domain(params)
        return params[indices]

    def clamp_params(self, params: Params) -> Params:
        raise NotImplementedError

    def __repr__(self):
        args = ', '.join(f'{k}={str(v)[:10]}' for k, v in vars(self).items())
        return f'{self.__class__.__name__}({args})'



from dataclasses import dataclass

import numpy as np

from reflectorch.inference.preprocess_exp.interpolation import interp_reflectivity
from reflectorch.inference.preprocess_exp.footprint import apply_footprint_correction, BEAM_SHAPE
from reflectorch.inference.preprocess_exp.normalize import intensity2reflectivity, NORMALIZE_MODE
from reflectorch.inference.preprocess_exp.attenuation import apply_attenuation_correction
from reflectorch.inference.preprocess_exp.cut_with_q_ratio import cut_curve
from reflectorch.utils import angle_to_q


def standard_preprocessing(
        intensity: np.ndarray,
        scattering_angle: np.ndarray,
        attenuation: np.ndarray,
        q_interp: np.ndarray,
        wavelength: float,
        beam_width: float,
        sample_length: float,
        min_intensity: float = 1e-10,
        beam_shape: BEAM_SHAPE = "gauss",
        normalize_mode: NORMALIZE_MODE = "max",
        incoming_intensity: float = None,
        max_q: float = None,  # if provided, max_angle is ignored
        max_angle: float = None,
) -> dict:
    """Preprocesses a raw experimental reflectivity curve by applying attenuation correction, footprint correction, cutting at a maximum q value and interpolation

    Args:
        intensity (np.ndarray): array of intensities of the reflectivity curve
        scattering_angle (np.ndarray): array of scattering angles
        attenuation (np.ndarray): attenuation factors for each measured point
        q_interp (np.ndarray): reciprocal space points used for the interpolation
        wavelength (float): the wavelength of the beam
        beam_width (float): the beam width
        sample_length (float): the sample length
        min_intensity (float, optional): intensities lower than this value are removed. Defaults to 1e-10.
        beam_shape (BEAM_SHAPE, optional): the shape of the beam, either "gauss" or "box". Defaults to "gauss".
        normalize_mode (NORMALIZE_MODE, optional): normalization mode, either "first", "max" or "incoming_intensity". Defaults to "max".
        incoming_intensity (float, optional): array of intensities for the "incoming_intensity" normalization. Defaults to None.
        max_q (float, optional): the maximum q value at which the curve is cut. Defaults to None.
        max_angle (float, optional): the maximum scattering angle at which the curve is cut; only used if max_q is not provided. Defaults to None.

    Returns:
        dict: dictionary containing the interpolated reflectivity curve, the curve before interpolation, the q values before interpolation, the q values after interpolation and the q ratio of the cutting
    """
    intensity = apply_attenuation_correction(
        intensity,
        attenuation,
        scattering_angle,
    )

    intensity = apply_footprint_correction(
        intensity, scattering_angle, beam_width=beam_width, sample_length=sample_length, beam_shape=beam_shape
    )

    curve = intensity2reflectivity(intensity, normalize_mode, incoming_intensity)

    curve, scattering_angle = remove_low_statistics(curve, scattering_angle, thresh=min_intensity)

    q = angle_to_q(scattering_angle, wavelength)

    q, curve, q_ratio = cut_curve(q, curve, max_q, max_angle, wavelength)

    curve_interp = interp_reflectivity(q_interp, q, curve)

    assert np.all(np.isfinite(curve_interp))
    assert np.all(np.isfinite(curve))
    assert np.all(np.isfinite(q))
    assert np.all(np.isfinite(q_interp))
    assert np.all(curve > 0.)
    assert np.all(curve_interp > 0.)

    return {
        "curve_interp": curve_interp, "curve": curve, "q_values": q, "q_interp": q_interp, "q_ratio": q_ratio,
    }


def remove_low_statistics(curve, scattering_angle, thresh: float = 1e-7):
    indices = (curve > thresh) & np.isfinite(curve)
    return curve[indices], scattering_angle[indices]


@dataclass
class StandardPreprocessing:
    q_interp: np.ndarray = None
    wavelength: float = 1.
    beam_width: float = None
    sample_length: float = None
    beam_shape: BEAM_SHAPE = "gauss"
    normalize_mode: NORMALIZE_MODE = "max"
    incoming_intensity: float = None

    def preprocess(self,
                   intensity: np.ndarray,
                   scattering_angle: np.ndarray,
                   attenuation: np.ndarray,
                   **kwargs
                   ) -> dict:
        attrs = self._get_updated_attrs(**kwargs)
        return standard_preprocessing(
            intensity,
            scattering_angle,
            attenuation,
            **attrs
        )

    __call__ = preprocess

    def set_parameters(self, **kwargs) -> None:
        for k, v in kwargs.items():
            if k in self.__annotations__:
                setattr(self, k, v)
            else:
                raise KeyError(f'Unknown parameter {k}.')

    def _get_updated_attrs(self, **kwargs):
        current_attrs = {k: getattr(self, k) for k in self.__annotations__.keys()}
        current_attrs.update(kwargs)
        return current_attrs



from typing import Union, Tuple
from math import log

import torch
from torch import Tensor

from reflectorch.data_generation.priors.base import PriorSampler
from reflectorch.data_generation.priors.params import Params
from reflectorch.data_generation.utils import get_param_labels

__all__ = [
    'SingleParamPrior',
    'UniformParamPrior',
    'GaussianParamPrior',
    'TruncatedGaussianParamPrior',
    'SimplePriorSampler',

]


class SingleParamPrior(object):
    def sample(self, batch_num: int, device=None, dtype=None):
        raise NotImplementedError

    def log_prob(self, params: Tensor):
        raise NotImplementedError

    def to_conf(self):
        vars_dict = {k: v for k, v in vars(self).items() if not k.startswith('_')}
        return {
            'cls': self.__class__.__name__,
            'kwargs': vars_dict
        }

    def scale(self, params: Tensor) -> Tensor:
        raise NotImplementedError

    def restore(self, params: Tensor) -> Tensor:
        raise NotImplementedError


class SimplePriorSampler(PriorSampler):
    def __init__(self,
                 *params: Union[Tuple[float, float], SingleParamPrior],
                 device=None, dtype=None, param_cls=None, scaling_prior: PriorSampler = None,
                 ):
        if param_cls is not None:
            self.PARAM_CLS = param_cls
        elif scaling_prior is not None:
            self.PARAM_CLS = scaling_prior.PARAM_CLS
        self.param_priors: Tuple[SingleParamPrior, ...] = self._init_params(params)
        self._num_layers = self.PARAM_CLS.size2layers_num(len(params))
        self.device = device
        self.dtype = dtype
        self.scaling_prior = scaling_prior

    @property
    def max_num_layers(self) -> int:
        return self._num_layers

    def _init_params(self, params: Tuple[Union[Tuple[float, float], SingleParamPrior], ...]):
        assert len(params) == self.PARAM_CLS.layers_num2size(self.PARAM_CLS.size2layers_num(len(params)))
        params = tuple(
            param if isinstance(param, SingleParamPrior) else UniformParamPrior(*param)
            for param in params
        )
        return params

    def sample(self, batch_size: int) -> Params:
        t_params = torch.stack([
            param.sample(batch_size, device=self.device, dtype=self.dtype) for param in self.param_priors
        ], -1)
        params = self.PARAM_CLS.from_tensor(t_params)
        return params

    def scale_params(self, params: Params) -> Tensor:
        if self.scaling_prior:
            return self.scaling_prior.scale_params(params)

        t_params = params.as_tensor()

        scaled_params = torch.stack(
            [param_prior.scale(param) for param, param_prior in zip(t_params.T, self.param_priors)], -1
        )

        return scaled_params

    def restore_params(self, scaled_params: Tensor) -> Params:
        if self.scaling_prior:
            return self.scaling_prior.restore_params(scaled_params)
        t_params = torch.stack(
            [param_prior.restore(param) for param, param_prior in zip(scaled_params.T, self.param_priors)], -1
        )
        return self.PARAM_CLS.from_tensor(t_params)

    def log_prob(self, params: Params) -> Tensor:
        t_params = params.as_tensor()
        log_probs = torch.stack(
            [param_prior.log_prob(param) for param, param_prior in zip(t_params.T, self.param_priors)], -1
        )
        return log_probs.sum(1)

    def get_indices_within_domain(self, params: Params) -> Tensor:
        log_probs = self.log_prob(params)
        return torch.isfinite(log_probs)

    def get_indices_within_bounds(self, params: Params) -> Tensor:
        return self.get_indices_within_domain(params)

    def __repr__(self):
        layers_num = self.PARAM_CLS.size2layers_num(len(self.param_priors))
        labels = get_param_labels(layers_num)
        prior_str = '\n\t'.join(f'{label}: {param_prior}' for label, param_prior in zip(labels, self.param_priors))
        return f'SimplePriorSampler(\n\t{prior_str}\n)'


class UniformParamPrior(SingleParamPrior):
    def __init__(self, min_value: float, max_value: float, device=None, dtype=None):
        assert min_value < max_value
        self.min_value, self.max_value, self.delta = min_value, max_value, max_value - min_value
        self._lob_prob_const = - log(self.delta)
        self.device = device
        self.dtype = dtype

    def sample(self, batch_num: int, device=None, dtype=None):
        params = torch.rand(
            batch_num, device=(device or self.device), dtype=(dtype or self.dtype)
        ) * self.delta + self.min_value
        return params

    def log_prob(self, params: Tensor):
        log_probs = torch.fill_(torch.ones_like(params), self._lob_prob_const)
        log_probs[(params < self.min_value) | (params > self.max_value)] = - float('inf')
        return log_probs

    def scale(self, params: Tensor) -> Tensor:
        return (params - self.min_value) / self.delta

    def restore(self, params: Tensor) -> Tensor:
        return params * self.delta + self.min_value

    def __repr__(self):
        return f'UniformParamPrior(min={self.min_value}, max={self.max_value})'


class GaussianParamPrior(SingleParamPrior):
    _GAUSS_SCALE_CONST: float = 4.

    def __init__(self, mean: float, std: float, device=None, dtype=None):
        assert std > 0
        self.mean = mean
        self.std = std
        self.device = device
        self.dtype = dtype

    def sample(self, batch_num: int, device=None, dtype=None):
        params = torch.normal(
            self.mean, self.std, (batch_num,), device=(device or self.device), dtype=(dtype or self.dtype)
        )
        return params

    def log_prob(self, params: Tensor):
        # ignore constant
        log_probs = - ((params - self.mean) / self.std) ** 2 / 2
        return log_probs

    def scale(self, params: Tensor) -> Tensor:
        return (params - self.mean) / (self.std * self._GAUSS_SCALE_CONST)

    def restore(self, params: Tensor) -> Tensor:
        return params * self.std * self._GAUSS_SCALE_CONST + self.mean

    def __repr__(self):
        return f'GaussianParamPrior(mean={self.mean}, std={self.std})'


class TruncatedGaussianParamPrior(GaussianParamPrior):
    def sample(self, batch_num: int, device=None, dtype=None) -> Tensor:
        params = torch.normal(
            self.mean, self.std, (batch_num,), device=(device or self.device), dtype=(dtype or self.dtype)
        )
        negative_params = params < 0.
        num_negative_params = negative_params.sum().item()
        if num_negative_params:
            params[negative_params] = self.sample(num_negative_params, device=device, dtype=dtype)
        return params

    def log_prob(self, params: Tensor) -> Tensor:
        # ignore constant
        log_probs: Tensor = - ((params - self.mean) / self.std) ** 2 / 2
        log_probs[params < 0.] = - float('inf')
        return log_probs

    def __repr__(self):
        return f'TruncatedGaussianParamPrior(mean={self.mean}, std={self.std})'



import numpy as np


def apply_attenuation_correction(
        intensity: np.ndarray,
        attenuation: np.ndarray,
        scattering_angle: np.ndarray = None,
        correct_discontinuities: bool = True
) -> np.ndarray:
    """Applies attenuation correction to experimental reflectivity curves

    Args:
        intensity (np.ndarray): intensities of an experimental reflectivity curve
        attenuation (np.ndarray): attenuation factors for each measured point
        scattering_angle (np.ndarray, optional): scattering angles of the measured points. Defaults to None.
        correct_discontinuities (bool, optional): whether to correct discontinuities in the measured curves. Defaults to True.

    Returns:
        np.ndarray: the corrected reflectivity curve
    """
    intensity = intensity / attenuation
    if correct_discontinuities:
        if scattering_angle is None:
            raise ValueError("correct_discontinuities options requires scattering_angle, but scattering_angle is None.")
        intensity = apply_discontinuities_correction(intensity, scattering_angle)
    return intensity


def apply_discontinuities_correction(intensity: np.ndarray, scattering_angle: np.ndarray) -> np.ndarray:
    intensity = intensity.copy()
    diff_angle = np.diff(scattering_angle)
    for i in range(len(diff_angle)):
        if diff_angle[i] == 0:
            factor = intensity[i] / intensity[i + 1]
            intensity[(i + 1):] *= factor
    return intensity



from __future__ import annotations
from typing import Union

import torch
from torch import nn, Tensor, stack, cat
from reflectorch.models.activations import activation_by_name
import reflectorch

###embedding network adapted from the PANPE repository

__all__ = [
    "IntegralConvEmbedding",
]

class IntegralConvEmbedding(nn.Module):
    def __init__(
        self,
        z_num: Union[int, tuple[int, ...]],
        z_range: tuple[float, float] = None,
        in_dim: int = 2,
        kernel_coef: int = 16,
        dim_embedding: int = 256,
        conv_dims: tuple[int, ...] = (32, 64, 128),
        num_blocks: int = 4,
        use_batch_norm: bool = False,
        use_layer_norm: bool = True,
        use_fft: bool = False,
        activation: str = "gelu",
        conv_activation: str = "lrelu",
        resnet_activation: str = "relu",
    ) -> None:
        super().__init__()

        if isinstance(z_num, int):
            z_num = (z_num,)
        num_kernel = len(z_num)

        if z_range is not None:
            zs = [(z_range[0], z_range[1], nz) for nz in z_num]
        else:
            zs = z_num

        self.in_dim = in_dim

        self.kernels = nn.ModuleList(
            [
                IntegralKernelBlock(
                    z,
                    in_dim,
                    kernel_coef=kernel_coef,
                    latent_dim=dim_embedding,
                    conv_dims=conv_dims,
                    use_fft=use_fft,
                    activation=activation,
                    conv_activation=conv_activation,
                )
                for z in zs
            ]
        )

        self.fc = reflectorch.models.networks.residual_net.ResidualMLP(
            dim_in=dim_embedding * num_kernel, 
            dim_out=dim_embedding,
            layer_width=2 * dim_embedding,
            num_blocks=num_blocks,
            use_batch_norm=use_batch_norm,
            use_layer_norm=use_layer_norm,
            activation=resnet_activation,
        )

    def forward(self, q, y, drop_mask=None) -> Tensor:
        x = cat([kernel(q, y, drop_mask=drop_mask) for kernel in self.kernels], dim=-1)
        x = self.fc(x)
        
        return x


class IntegralKernelBlock(nn.Module):
    """
    Examples:
        >>> x = torch.rand(2, 100)
        >>> y = torch.rand(2, 100, 3)
        >>> block = IntegralKernelBlock((0, 1, 10), in_dim=3,  latent_dim=32)
        >>> output = block(x, y)
        >>> output.shape
        torch.Size([2, 32])

        >>> block = IntegralKernelBlock(10, in_dim=3,  latent_dim=32)
        >>> output = block(x, y)
        >>> output.shape
        torch.Size([2, 32])
    """

    def __init__(
        self,
        z: tuple[float, float, int] or int,
        in_dim: int,
        kernel_coef: int = 2,
        latent_dim: int = 32,
        conv_dims: tuple[int, ...] = (32, 64, 128),
        use_fft: bool = False,
        activation: str = "gelu",
        conv_activation: str = "lrelu",
    ):
        super().__init__()

        if isinstance(z, int):
            z_num = z
            kernel = FullIntegralKernel(z_num, in_dim=in_dim, kernel_coef=kernel_coef)
        else:
            kernel = FastIntegralKernel(
                z, in_dim=in_dim, kernel_coef=kernel_coef, activation=activation
            )
            z_num = z[-1]

        assert z_num % 2 == 0, "z_num should be even"

        self.kernel = kernel
        self.z_num = z_num
        self.in_dim = in_dim
        self.latent_dim = latent_dim
        self.use_fft = use_fft

        self.fc_in_dim = self.latent_dim + self.in_dim * self.z_num
        if self.use_fft:
            self.fc_in_dim += self.in_dim * 2 + self.in_dim * self.z_num

        self.conv = reflectorch.models.encoders.conv_encoder.ConvEncoder(
            dim_avpool=8,
            hidden_channels=conv_dims,
            in_channels=in_dim,
            dim_embedding=latent_dim,
            activation=conv_activation,
        )
        self.fc = FCBlock(
            in_dim=self.fc_in_dim, hid_dim=self.latent_dim * 2, out_dim=self.latent_dim
        )

    def forward(self, x: Tensor, y: Tensor, drop_mask: Tensor = None) -> Tensor:
        x = self.kernel(x, y, drop_mask=drop_mask)

        assert x.shape == (x.shape[0], self.in_dim, self.z_num)

        xc = self.conv(x)  # (batch, latent_dim)

        assert xc.shape == (x.shape[0], self.latent_dim)

        if self.use_fft:
            fft_x = torch.fft.rfft(x, dim=-1, norm="ortho")  # (batch, in_dim, z_num)

            fft_x = torch.cat(
                [fft_x.real, fft_x.imag], -1
            )  # (batch, in_dim, 2 * z_num)

            assert fft_x.shape == (x.shape[0], x.shape[1], self.z_num + 2)

            fft_x = fft_x.flatten(1)  # (batch, in_dim * (z_num + 2))

            x = torch.cat(
                [x.flatten(1), fft_x, xc], -1
            )  # (batch, in_dim * z_num * 3 + latent_dim)
        else:
            x = torch.cat([x.flatten(1), xc], -1)

        assert (
            x.shape[1] == self.fc_in_dim
        ), f"Expected dim {self.fc_in_dim}, got {x.shape[1]}"

        x = self.fc(x)  # (batch, latent_dim)

        return x


class FastIntegralKernel(nn.Module):
    def __init__(
        self,
        z: tuple[float, float, int],
        kernel_coef: int = 16,
        in_dim: int = 1,
        activation: str = "gelu",
    ):
        super().__init__()

        z = torch.linspace(*z)

        self.kernel = FCBlock(
            in_dim + 2, kernel_coef * in_dim, in_dim, activation=activation
        )

        self.register_buffer("z", z)

    def _get_z(self, x: Tensor):
        # x.shape == (batch_size, num_x)
        dz = self.z[1] - self.z[0]
        indices = torch.ceil((x - self.z[0] - dz / 2) / dz).to(torch.int64)

        z = torch.index_select(self.z, 0, indices.flatten()).view(*x.shape)

        return z, indices

    def forward(self, x: Tensor, y: Tensor, drop_mask=None):
        z, indices = self._get_z(x)
        xz = torch.stack([x, z], -1)
        kernel_input = torch.cat([xz, y], -1)
        output = self.kernel(kernel_input)  # (batch, x_num, in_dim)

        output = compute_means(
            output * y, indices, self.z.shape[-1], drop_mask=drop_mask
        )  # (batch, z_num, in_dim)

        output = output.swapaxes(1, 2)  # (batch, in_dim, z_num)

        return output


class FullIntegralKernel(nn.Module):
    def __init__(
        self,
        z_num: int,
        kernel_coef: int = 1,
        in_dim: int = 1,
    ):
        super().__init__()

        self.z_num = z_num
        self.in_dim = in_dim

        self.kernel = nn.Sequential(
            nn.Linear(in_dim + 1, z_num * kernel_coef),
            nn.LayerNorm(z_num * kernel_coef),
            nn.ReLU(),
            nn.Linear(z_num * kernel_coef, z_num * in_dim),
        )

    def forward(self, x: Tensor, y: Tensor, drop_mask=None):
        # x.shape == (batch_size, num_x)
        # y.shape == (batch_size, num_x, in_dim)
        # drop_mask.shape == (batch_size, num_x)

        batch_size, num_x = x.shape

        kernel_input = torch.cat([x.unsqueeze(-1), y], -1)  # (batch, x_num, in_dim + 1)
        x = self.kernel(kernel_input)  # (batch, x_num, z_num * in_dim)
        x = x.reshape(
            *x.shape[:-1], self.z_num, self.in_dim
        )  # (batch, x_num, z_num, in_dim)
        # permute to get (batch, z_num, x_num, in_dim)
        x = x.permute(0, 2, 1, 3)

        y = y.unsqueeze(1)  # (batch, 1, x_num, in_dim)

        assert x.shape == (
            batch_size,
            self.z_num,
            num_x,
            self.in_dim,
        )  # (batch, z_num, in_dim, x_num)
        assert y.shape == (
            batch_size,
            1,
            num_x,
            self.in_dim,
        )  # (batch, 1, x_num, in_dim)

        if drop_mask is not None:
            x = x * y
            x = x.permute(0, 2, 1, 3)  # (batch, x_num, z_num, in_dim)
            x = masked_mean(x, drop_mask)
        else:
            x = (x * y).mean(-2)  # (batch, z_num, in_dim)

        assert x.shape == (batch_size, self.z_num, self.in_dim), f"{x.shape}"

        x = x.swapaxes(1, 2)  # (batch, in_dim, z_num)

        return x


class FCBlock(nn.Module):
    def __init__(
        self,
        in_dim: int = 2,
        hid_dim: int = 16,
        out_dim: int = 16,
        activation: str = "gelu",
    ):
        super().__init__()

        self.fc1 = nn.Linear(in_dim, hid_dim)
        self.layer_norm = nn.LayerNorm(hid_dim)
        self.activation = activation_by_name(activation)()
        self.fc2 = nn.Linear(hid_dim, out_dim)

    def forward(self, x: Tensor) -> Tensor:
        x = self.fc1(x)
        x = self.layer_norm(x)
        x = self.activation(x)
        x = self.fc2(x)
        return x
        # return self.kernel(x)


def compute_means(x, indices, z: int, drop_mask: Tensor = None):
    """
    Compute the mean values of tensor 'x' for each unique index in 'indices' across each batch.

    This function calculates the mean of elements in 'x' that correspond to each unique index in 'indices'.
    The computation is performed for each batch separately, and the function is optimized to avoid Python loops
    by using advanced PyTorch operations.

    Parameters:
    x (torch.Tensor): A tensor of shape (batch_size, n, d) containing the values to be averaged.
                      'x' should be a floating-point tensor.
    indices (torch.Tensor): An integer tensor of shape (batch_size, n) containing the indices.
                            The values in 'indices' should be in the range [0, z-1].
    z (int): The number of unique indices. This determines the second dimension of the output tensor.
    drop_mask (torch.Tensor): A boolean tensor of shape (batch_size, n) containing a mask for the indices to drop.
                              If None, all indices are used.

    Returns:
    torch.Tensor: A tensor of shape (batch_size, z, d) containing the mean values for each index in each batch.
                  If an index does not appear in a batch, its corresponding mean values are zeros.

    Example:
    >>> batch_size, n, d, z = 3, 4, 5, 6
    >>> indices = torch.randint(0, z, (batch_size, n))
    >>> x = torch.randn(batch_size, n, d)
    >>> y = compute_means(x, indices, z)
    >>> print(y.shape)
    torch.Size([3, 6, 5])
    """

    batch_size, n, d = x.shape
    device = x.device

    drop = drop_mask is not None

    # Initialize tensors to hold sums and counts
    sums = torch.zeros(batch_size, z + int(drop), d, device=device)
    counts = torch.zeros(batch_size, z + int(drop), device=device)

    if drop_mask is not None:
        # Set the values of the indices to drop to z
        indices = indices.masked_fill(~drop_mask, z)

    indices_expanded = indices.unsqueeze(-1).expand_as(x)
    sums.scatter_add_(1, indices_expanded, x)
    counts.scatter_add_(1, indices, torch.ones_like(indices, dtype=x.dtype))

    if drop:
        # Remove the z values from the sums and counts
        sums = sums[:, :-1]
        counts = counts[:, :-1]

    # Compute the mean and handle division by zero
    mean = sums / counts.unsqueeze(-1).clamp(min=1)

    return mean


def masked_mean(x, mask):
    """
    Computes the mean of tensor x along the x_size dimension,
    while masking out elements where the corresponding value in the mask is False.

    Args:
    x (torch.Tensor): A tensor of shape (batch, x_size, z, d).
    mask (torch.Tensor): A boolean mask of shape (batch, x_size).

    Returns:
    torch.Tensor: The result tensor of shape (batch, z, d) after applying the mask and computing the mean.
    """
    if not mask.dtype == torch.bool:
        raise TypeError("Mask must be a boolean tensor.")

    # Ensure the mask is broadcastable to the shape of x
    mask = mask.unsqueeze(-1).unsqueeze(-1)
    masked_x = x * mask

    # Compute the sum and the count of valid (unmasked) elements along the x_size dimension
    sum_x = masked_x.sum(dim=1)
    count_x = mask.sum(dim=1)

    # Avoid division by zero
    count_x[count_x == 0] = 1

    # Compute the mean
    mean_x = sum_x / count_x

    return mean_x


from typing import Union, Tuple

import torch
from torch import Tensor

from reflectorch.data_generation import (
    PriorSampler,
    Params,
)


class LogLikelihood(object):
    """Computes the gaussian log likelihood of the thin film parameters

    Args:
        q (Tensor): the q values
        exp_curve (Tensor): the experimental reflectivity curve
        priors (PriorSampler): the prior sampler
        sigmas (Union[float, Tensor]): the sigmas (i.e. intensity error bars)
    """
    def __init__(self, q: Tensor, exp_curve: Tensor, priors: PriorSampler, sigmas: Union[float, Tensor]):
        self.exp_curve = torch.atleast_2d(exp_curve)
        self.priors: PriorSampler = priors
        self.q = q
        self.sigmas = sigmas
        self.sigmas2 = self.sigmas ** 2

    def calc_log_likelihood(self, curves: Tensor):
        "computes the gaussian log likelihood"
        log_probs = - (self.exp_curve - curves) ** 2 / self.sigmas2 / 2
        return log_probs.sum(-1)

    def __call__(self, params: Union[Params, Tensor], curves: Tensor = None):
        if not isinstance(params, Params):
            params: Params = self.priors.PARAM_CLS.from_tensor(params)
        log_priors: Tensor = self.priors.log_prob(params)
        indices: Tensor = torch.isfinite(log_priors)

        if not indices.sum().item():
            return log_priors

        finite_params: Params = params[indices]

        if curves is None:
            curves: Tensor = finite_params.reflectivity(self.q)
        else:
            curves = curves[indices]

        log_priors[indices] += self.calc_log_likelihood(curves)

        return log_priors

    calc_log_posterior = __call__

    def get_importance_sampling_weights(
            self, sampled_params: Params, nf_log_probs: Tensor, curves: Tensor = None
    ) -> Tuple[Tensor, Tensor, Tensor]:
        log_probs = self.calc_log_posterior(sampled_params, curves=curves)
        log_weights = log_probs - nf_log_probs
        log_weights = log_weights - log_weights.max()

        weights = torch.exp(log_weights.to(torch.float64)).to(log_weights)
        weights = weights / weights.sum()

        return weights, log_weights, log_probs


class PoissonLogLikelihood(LogLikelihood):
    """Computes the Poisson log likelihood of the thin film parameters

    Args:
        q (Tensor): the q values
        exp_curve (Tensor): the experimental reflectivity curve
        priors (PriorSampler): the prior sampler
        sigmas (Union[float, Tensor]): the sigmas (i.e. intensity error bars)
    """    
    def calc_log_likelihood(self, curves: Tensor):
        """computes the Poisson log likelihood"""
        log_probs = self.exp_curve / self.sigmas2 * (self.exp_curve * torch.log(curves) - curves)
        return log_probs.sum(-1)



from functools import lru_cache
from typing import Tuple

import torch
from torch import Tensor

from reflectorch.data_generation.utils import (
    uniform_sampler,
    logdist_sampler,
    triangular_sampler,
    get_slds_from_d_rhos,
)

from reflectorch.data_generation.priors.params import Params
from reflectorch.data_generation.priors.no_constraints import (
    BasicPriorSampler,
    DEFAULT_ROUGHNESS_RANGE,
    DEFAULT_THICKNESS_RANGE,
    DEFAULT_SLD_RANGE,
    DEFAULT_NUM_LAYERS,
    DEFAULT_DEVICE,
    DEFAULT_DTYPE,
    DEFAULT_SCALED_RANGE,
    DEFAULT_USE_DRHO,
)


class UniformSubPriorParams(Params):
    """Parameters class for thicknesses, roughnesses and slds, together with their subprior bounds."""
    __slots__ = ('thicknesses', 'roughnesses', 'slds', 'min_bounds', 'max_bounds')
    PARAM_NAMES = __slots__

    def __init__(self,
                 thicknesses: Tensor,
                 roughnesses: Tensor,
                 slds: Tensor,
                 min_bounds: Tensor,
                 max_bounds: Tensor,
                 ):
        super().__init__(thicknesses, roughnesses, slds)
        self.min_bounds = min_bounds
        self.max_bounds = max_bounds

    @staticmethod
    def rearrange_context_from_params(
            scaled_params: Tensor, context: Tensor, inference: bool = False, from_params: bool = False
    ):
        if inference:
            if from_params:
                num_params = scaled_params.shape[1] // 3
                scaled_params = scaled_params[:, num_params:]
            context = torch.cat([context, scaled_params], dim=-1)
            return context

        num_params = scaled_params.shape[1] // 3
        assert num_params * 3 == scaled_params.shape[1]
        scaled_params, bound_context = torch.split(scaled_params, [num_params, 2 * num_params], dim=-1)
        context = torch.cat([context, bound_context], dim=-1)
        return scaled_params, context

    @staticmethod
    def restore_params_from_context(scaled_params: Tensor, context: Tensor):
        num_params = scaled_params.shape[-1]
        scaled_bounds = context[:, -2 * num_params:]
        scaled_params = torch.cat([scaled_params, scaled_bounds], dim=-1)
        return scaled_params

    @staticmethod
    def input_context_split(t_params):
        num_params = t_params.shape[1] // 3
        return torch.split(t_params, [num_params, 2 * num_params])

    def as_tensor(self, use_drho: bool = False, add_bounds: bool = True) -> Tensor:
        t_list = [self.thicknesses, self.roughnesses]
        if use_drho:
            t_list.append(self.d_rhos)
        else:
            t_list.append(self.slds)
        if add_bounds:
            t_list += [self.min_bounds, self.max_bounds]
        return torch.cat(t_list, -1)

    @classmethod
    def from_tensor(cls, params: Tensor):
        layers_num = (params.shape[-1] - 6) // 9
        num_params = 3 * layers_num + 2

        thicknesses, roughnesses, slds, min_bounds, max_bounds = torch.split(
            params,
            [layers_num, layers_num + 1, layers_num + 1, num_params, num_params],
            dim=-1
        )

        return cls(thicknesses, roughnesses, slds, min_bounds, max_bounds)

    @property
    def num_params(self) -> int:
        return self.layers_num2size(self.max_layer_num)

    @staticmethod
    def size2layers_num(size: int) -> int:
        return (size - 6) // 9

    @staticmethod
    def layers_num2size(layers_num: int) -> int:
        return layers_num * 9 + 6

    def scale_with_q(self, q_ratio: float):
        super().scale_with_q(q_ratio)

        layer_num = self.max_layer_num
        scales = torch.tensor(
            [1 / q_ratio] * (2 * layer_num + 1) + [q_ratio ** 2] * (layer_num + 1),
            device=self.device, dtype=self.dtype
        )

        self.min_bounds *= scales
        self.max_bounds *= scales


class UniformSubPriorSampler(BasicPriorSampler):
    """Prior sampler for thicknesses, roughnesses, slds and their subprior bounds

    Args:
        thickness_range (Tuple[float, float], optional): the range of the layer thicknesses. Defaults to DEFAULT_THICKNESS_RANGE.
        roughness_range (Tuple[float, float], optional): the range of the interlayer roughnesses. Defaults to DEFAULT_ROUGHNESS_RANGE.
        sld_range (Tuple[float, float], optional): the range of the layer SLDs. Defaults to DEFAULT_SLD_RANGE.
        num_layers (int, optional): the number of layers. Defaults to DEFAULT_NUM_LAYERS.
        use_drho (bool, optional): whether to use differences in SLD values between neighboring layers instead of the actual SLD values. Defaults to DEFAULT_USE_DRHO.
        device (torch.device, optional): the Pytorch device. Defaults to DEFAULT_DEVICE.
        dtype (torch.dtype, optional): the Pytorch data type. Defaults to DEFAULT_DTYPE.
        scaled_range (Tuple[float, float], optional): the range for scaling the parameters. Defaults to DEFAULT_SCALED_RANGE.
        scale_by_subpriors (bool, optional): if True the film parameters are scaled with respect to their subprior bounds. Defaults to False.
        smaller_roughnesses (bool, optional): if True the sampled roughnesses are biased towards smaller values. Defaults to False.
        logdist (bool, optional): if True the relative widths of the subprior intervals are sampled uniformly on a logarithmic scale instead of uniformly. Defaults to False.
        relative_min_bound_width (float, optional): defines the interval [relative_min_bound_width, 1.0] from which the relative bound widths for each parameter are sampled. Defaults to 1e-2.
    """
    PARAM_CLS = UniformSubPriorParams

    def __init__(self,
                 thickness_range: Tuple[float, float] = DEFAULT_THICKNESS_RANGE,
                 roughness_range: Tuple[float, float] = DEFAULT_ROUGHNESS_RANGE,
                 sld_range: Tuple[float, float] = DEFAULT_SLD_RANGE,
                 num_layers: int = DEFAULT_NUM_LAYERS,
                 use_drho: bool = DEFAULT_USE_DRHO,
                 device: torch.device = DEFAULT_DEVICE,
                 dtype: torch.dtype = DEFAULT_DTYPE,
                 scaled_range: Tuple[float, float] = DEFAULT_SCALED_RANGE,
                 scale_by_subpriors: bool = False,
                 smaller_roughnesses: bool = False,
                 logdist: bool = False,
                 relative_min_bound_width: float = 1e-2,
                 ):
        super().__init__(
            thickness_range,
            roughness_range,
            sld_range,
            num_layers,
            use_drho,
            device,
            dtype,
            scaled_range,
        )

        self.scale_by_subpriors = scale_by_subpriors
        self.smaller_roughnesses = smaller_roughnesses
        self.logdist = logdist
        self.relative_min_bound_width = relative_min_bound_width

    @property
    def max_num_layers(self) -> int:
        return self.num_layers

    @property
    def param_dim(self) -> int:
        return self.max_num_layers * 3 + 2

    @lru_cache()
    def min_vector(self, layers_num, drho: bool = False):
        min_vector = super().min_vector(layers_num, drho)
        min_vector = torch.cat([min_vector, min_vector, min_vector], dim=0)
        return min_vector

    def scale_params(self, params: UniformSubPriorParams) -> Tensor:
        scaled_params = super().scale_params(params)

        if self.scale_by_subpriors:
            params_t = params.as_tensor(use_drho=self.use_drho, add_bounds=False)
            scaled_params[:, :self.param_dim] = self._scale(params_t, params.min_bounds, params.max_bounds)

        return scaled_params

    def restore_params(self, scaled_params: Tensor) -> Params:
        if not self.scale_by_subpriors:
            return super().restore_params(scaled_params)

        scaled_params, scaled_min_bounds, scaled_max_bounds = torch.split(
            scaled_params, [self.param_dim, self.param_dim, self.param_dim], dim=1
        )

        min_vector = super().min_vector(self.max_num_layers, self.use_drho)
        max_vector = super().max_vector(self.max_num_layers, self.use_drho)

        min_bounds = self._restore(scaled_min_bounds, min_vector, max_vector)
        max_bounds = self._restore(scaled_max_bounds, min_vector, max_vector)

        param_t = self._restore(scaled_params, min_bounds, max_bounds)
        param_t = torch.cat([param_t, min_bounds, max_bounds], dim=-1)

        params = UniformSubPriorParams.from_tensor(param_t)

        if self.use_drho:
            params.slds = get_slds_from_d_rhos(params.slds)
        return params

    @lru_cache()
    def max_vector(self, layers_num, drho: bool = False):
        max_vector = super().max_vector(layers_num, drho)
        max_vector = torch.cat([max_vector, max_vector, max_vector], dim=0)
        return max_vector

    @lru_cache()
    def delta_vector(self, layers_num, drho: bool = False):
        delta_vector = self.max_vector(layers_num, drho) - self.min_vector(layers_num, drho)
        delta_vector[delta_vector == 0.] = 1.
        return delta_vector

    def get_indices_within_bounds(self, params: UniformSubPriorParams) -> Tensor:
        t_params = torch.cat([
            params.thicknesses,
            params.roughnesses,
            params.slds
        ], dim=-1)

        indices = (
                torch.all(t_params >= params.min_bounds, dim=-1) &
                torch.all(t_params <= params.max_bounds, dim=-1)
        )

        return indices

    def clamp_params(self, params: UniformSubPriorParams) -> UniformSubPriorParams:
        params = UniformSubPriorParams.from_tensor(
            torch.cat([
                torch.clamp(
                    params.as_tensor(add_bounds=False),
                    params.min_bounds, params.max_bounds
                ),
                params.min_bounds, params.max_bounds
            ], dim=1)
        )
        return params

    def get_indices_within_domain(self, params: UniformSubPriorParams) -> Tensor:
        return self.get_indices_within_bounds(params)

    def sample(self, batch_size: int) -> UniformSubPriorParams:
        min_bounds, max_bounds = self.sample_bounds(batch_size)

        params = torch.rand(
            *min_bounds.shape,
            device=self.device,
            dtype=self.dtype
        ) * (max_bounds - min_bounds) + min_bounds

        thicknesses, roughnesses, slds = torch.split(
            params, [self.max_num_layers, self.max_num_layers + 1, self.max_num_layers + 1], dim=-1
        )

        params = UniformSubPriorParams(thicknesses, roughnesses, slds, min_bounds, max_bounds)

        return params

    def sample_bounds(self, batch_size: int):
        min_vector, max_vector = (
            super().min_vector(self.num_layers)[None],
            super().max_vector(self.num_layers)[None]
        )

        delta_vector = max_vector - min_vector

        if self.logdist:
            widths_sampler_func = logdist_sampler
        else:
            widths_sampler_func = uniform_sampler

        prior_widths = widths_sampler_func(
            self.relative_min_bound_width, 1.,
            batch_size, delta_vector.shape[1],
            device=self.device, dtype=self.dtype
        ) * delta_vector

        prior_centers = uniform_sampler(
            min_vector + prior_widths / 2, max_vector - prior_widths / 2,
            *prior_widths.shape,
            device=self.device, dtype=self.dtype
        )

        if self.smaller_roughnesses:
            idx_min, idx_max = self.num_layers, self.num_layers * 2 + 1
            prior_centers[:, idx_min:idx_max] = triangular_sampler(
                min_vector[:, idx_min:idx_max] + prior_widths[:, idx_min:idx_max] / 2,
                max_vector[:, idx_min:idx_max] - prior_widths[:, idx_min:idx_max] / 2,
                batch_size, self.num_layers + 1,
                device=self.device, dtype=self.dtype
            )

        min_bounds, max_bounds = prior_centers - prior_widths / 2, prior_centers + prior_widths / 2

        return min_bounds, max_bounds

    def scale_bounds(self, bounds: Tensor) -> Tensor:
        layers_num = bounds.shape[-1] // 2

        return self._scale(
            bounds,
            self.min_vector(layers_num, drho=self.use_drho).to(bounds),
            self.max_vector(layers_num, drho=self.use_drho).to(bounds),
        )


class NarrowSldUniformSubPriorSampler(UniformSubPriorSampler):
    """Prior sampler for thicknesses, roughnesses, slds and their subprior bounds. The subprior bound widths for SLDs are restricted to be lower than a specified value. """
    def __init__(self,
                 thickness_range: Tuple[float, float] = DEFAULT_THICKNESS_RANGE,
                 roughness_range: Tuple[float, float] = DEFAULT_ROUGHNESS_RANGE,
                 sld_range: Tuple[float, float] = DEFAULT_SLD_RANGE,
                 num_layers: int = DEFAULT_NUM_LAYERS,
                 use_drho: bool = DEFAULT_USE_DRHO,
                 device: torch.device = DEFAULT_DEVICE,
                 dtype: torch.dtype = DEFAULT_DTYPE,
                 scaled_range: Tuple[float, float] = DEFAULT_SCALED_RANGE,
                 scale_by_subpriors: bool = False,
                 max_sld_prior_width: float = 10.,
                 ):
        super().__init__(
            thickness_range,
            roughness_range,
            sld_range,
            num_layers,
            use_drho,
            device,
            dtype,
            scaled_range,
            scale_by_subpriors,
        )

        self.max_sld_prior_width = max_sld_prior_width

    def sample_bounds(self, batch_size: int):
        min_vector, max_vector = (
            BasicPriorSampler.min_vector(self, self.num_layers),
            BasicPriorSampler.max_vector(self, self.num_layers),
        )

        delta_vector = max_vector - min_vector
        delta_vector[-self.num_layers:] = self.max_sld_prior_width

        prior_widths = uniform_sampler(
            delta_vector * self.relative_min_bound_width, delta_vector,
            batch_size, min_vector.shape[0],
            device=self.device, dtype=self.dtype
        )

        prior_centers = uniform_sampler(
            min_vector + prior_widths / 2, max_vector - prior_widths / 2,
            *prior_widths.shape,
            device=self.device, dtype=self.dtype
        )

        return prior_centers - prior_widths / 2, prior_centers + prior_widths / 2



from typing import Tuple, Union

import numpy as np

import torch
from torch import Tensor

from reflectorch.data_generation.utils import uniform_sampler
from reflectorch.data_generation.priors import BasicParams
from reflectorch.utils import angle_to_q
from reflectorch.data_generation.priors.no_constraints import DEFAULT_DEVICE, DEFAULT_DTYPE

__all__ = [
    "QGenerator",
    "ConstantQ",
    "VariableQ",
    "EquidistantQ",
    "ConstantAngle",
    "MaskedVariableQ",
]


class QGenerator(object):
    """Base class for momentum transfer (q) generators"""
    def get_batch(self, batch_size: int, context: dict = None) -> Tensor:
        pass


class ConstantQ(QGenerator):
    """Q generator for reflectivity curves with fixed discretization

    Args:
        q (Union[Tensor, Tuple[float, float, int]], optional): tuple (q_min, q_max, num_q) defining the minimum q value, maximum q value and the number of q points. Defaults to (0., 0.2, 128).
        device (optional): the Pytorch device. Defaults to DEFAULT_DEVICE.
        dtype (optional): the Pytorch data type. Defaults to DEFAULT_DTYPE.
        remove_zero (bool, optional): do not include the upper end of the interval. Defaults to False.
        fixed_zero (bool, optional): do not include the lower end of the interval. Defaults to False.
    """

    def __init__(self,
                 q: Union[Tensor, Tuple[float, float, int]] = (0., 0.2, 128),
                 device=DEFAULT_DEVICE,
                 dtype=DEFAULT_DTYPE,
                 remove_zero: bool = False,
                 fixed_zero: bool = False,
                 ):
        if isinstance(q, (tuple, list)):
            q = torch.linspace(*q, device=device, dtype=dtype)
            if remove_zero:
                if fixed_zero:
                    q = q[1:]
                else:
                    q = q[:-1]
        self.q_min = q.min().item()
        self.q_max = q.max().item()
        self.q = q

    def get_batch(self, batch_size: int, context: dict = None) -> Tensor:
        """generate a batch of q values

        Args:
            batch_size (int): the batch size

        Returns:
            Tensor: generated batch of q values
        """
        return self.q.clone()[None].expand(batch_size, self.q.shape[0])
    
    def scale_q(self, q):
        """Scales the q values to the range [-1, 1].

        Args:
            q (Tensor): unscaled q values

        Returns:
            Tensor: scaled q values
        """
        scaled_q_01 = (q - self.q_min) / (self.q_max - self.q_min)
        return 2.0 * (scaled_q_01 - 0.5)
    

class VariableQ(QGenerator):
    """Q generator for reflectivity curves with variable discretization

    Args:
        q_min_range (list, optional): the range for sampling the minimum q value of the curves, q_min. Defaults to [0.01, 0.03].
        q_max_range (list, optional): the range for sampling the maximum q value of the curves, q_max. Defaults to [0.1, 0.5].
        n_q_range (list, optional): the range for the number of points in the curves (equidistantly sampled between q_min and q_max, 
                                    the number of points varies between batches but is constant within a batch). Defaults to [64, 256].
        device (optional): the Pytorch device. Defaults to DEFAULT_DEVICE.
        dtype (optional): the Pytorch data type. Defaults to DEFAULT_DTYPE.
    """

    def __init__(self,
                 q_min_range: Tuple[float, float] = (0.01, 0.03),
                 q_max_range: Tuple[float, float] = (0.1, 0.5),
                 n_q_range: Tuple[int, int] = (64, 256),
                 mode: str = 'equidistant',
                 device=DEFAULT_DEVICE,
                 dtype=DEFAULT_DTYPE,
                 ):
        self.q_min_range = q_min_range
        self.q_max_range = q_max_range
        self.n_q_range = n_q_range
        self.mode = mode
        self.device = device
        self.dtype = dtype

    def get_batch(self, batch_size: int, context: dict = None) -> Tensor:
        """generate a batch of q values (the number of points varies between batches but is constant within a batch)

        Args:
            batch_size (int): the batch size

        Returns:
            Tensor: generated batch of q values
        """

        q_min = torch.rand(batch_size, device=self.device, dtype=self.dtype) * (self.q_min_range[1] - self.q_min_range[0]) + self.q_min_range[0]
        q_max = torch.rand(batch_size, device=self.device, dtype=self.dtype) * (self.q_max_range[1] - self.q_max_range[0]) + self.q_max_range[0]

        n_q = torch.randint(self.n_q_range[0], self.n_q_range[1] + 1, (1,), device=self.device).item()

        if self.mode == 'equidistant':
            q = torch.linspace(0, 1, n_q, device=self.device, dtype=self.dtype)
        elif self.mode == 'random':
            q = torch.rand(n_q, device=self.device, dtype=self.dtype).sort().values
        elif self.mode == 'logspace':
            q = torch.logspace(
                start=torch.log10(torch.tensor(1e-4, dtype=self.dtype, device=self.device)),
                end=torch.log10(torch.tensor(1.0, dtype=self.dtype, device=self.device)),
                steps=n_q, dtype=self.dtype, device=self.device)

        q = q_min[:, None] + q * (q_max - q_min)[:, None]
        
        return q
    
    def scale_q(self, q):
        """scales the q values to the range [-1, 1]

        Args:
            q (Tensor): unscaled q values

        Returns:
            Tensor: scaled q values
        """
        scaled_q_01 = (q - self.q_min_range[0]) / (self.q_max_range[1] - self.q_min_range[0]) 

        return 2.0 * (scaled_q_01 - 0.5)


class ConstantAngle(QGenerator):
    """Q generator for reflectivity curves measured at equidistant angles

    Args:
        angle_range (Tuple[float, float, int], optional): the range of the incident angles. Defaults to (0., 0.2, 257).
        wavelength (float, optional): the beam wavelength in units of angstroms. Defaults to 1.
        device (optional): the Pytorch device. Defaults to DEFAULT_DEVICE.
        dtype (optional): the Pytorch data type. Defaults to DEFAULT_DTYPE.
        """
    def __init__(self,
                 angle_range: Tuple[float, float, int] = (0., 0.2, 257),
                 wavelength: float = 1.,
                 device=DEFAULT_DEVICE,
                 dtype=DEFAULT_DTYPE,
                 ):
        self.q = torch.from_numpy(angle_to_q(np.linspace(*angle_range), wavelength)).to(device).to(dtype)

    def get_batch(self, batch_size: int, context: dict = None) -> Tensor:
        """generate a batch of q values

        Args:
            batch_size (int): the batch size

        Returns:
            Tensor: generated batch of q values
        """
        return self.q.clone()[None].expand(batch_size, self.q.shape[0])


class EquidistantQ(QGenerator):
    def __init__(self,
                 max_range: Tuple[float, float],
                 num_values: Union[int, Tuple[int, int]],
                 device=None,
                 dtype=torch.float64
                 ):
        self.max_range = max_range
        self._num_values = num_values
        self.device = device
        self.dtype = dtype

    @property
    def num_values(self) -> int:
        if isinstance(self._num_values, int):
            return self._num_values
        return np.random.randint(*self._num_values)

    def get_batch(self, batch_size: int, context: dict = None) -> Tensor:
        num_values = self.num_values
        q_max = uniform_sampler(*self.max_range, batch_size, 1, device=self.device, dtype=self.dtype)
        norm_qs = torch.linspace(0, 1, num_values + 1, device=self.device, dtype=self.dtype)[1:][None]
        qs = norm_qs * q_max
        return qs


class MaskedVariableQ:
    def __init__(self,
                 q_min_range=(0.01, 0.03),
                 q_max_range=(0.1, 0.5),
                 n_q_range=(64, 256),
                 mode='equidistant',
                 shuffle_mask=False,
                 total_thickness_constraint=True,
                 min_points_per_fringe=4,
                 device=DEFAULT_DEVICE,
                 dtype=DEFAULT_DTYPE):
        self.q_min_range = q_min_range
        self.q_max_range = q_max_range
        self.n_q_range = n_q_range
        self.device = device
        self.dtype = dtype
        self.mode = mode
        self.shuffle_mask = shuffle_mask
        self.total_thickness_constraint = total_thickness_constraint
        self.min_points_per_fringe = min_points_per_fringe
    
    def get_batch(self, batch_size, context):
        assert context is not None

        q_min = torch.rand(batch_size, device=self.device, dtype=self.dtype) * (self.q_min_range[1] - self.q_min_range[0]) + self.q_min_range[0]
        q_max = torch.rand(batch_size, device=self.device, dtype=self.dtype) * (self.q_max_range[1] - self.q_max_range[0]) + self.q_max_range[0]

        max_n_q = self.n_q_range[1]

        if self.mode == 'equidistant':
            positions = torch.linspace(0, 1, max_n_q, device=self.device, dtype=self.dtype).expand(batch_size, max_n_q)
        elif self.mode == 'random':
            positions = torch.rand(batch_size, max_n_q, device=self.device, dtype=self.dtype)
            positions, _ = positions.sort(dim=-1)
        elif self.mode == 'mixed':
            positions = torch.empty(batch_size, max_n_q, device=self.device, dtype=self.dtype)

            half = batch_size // 2 # half batch gets equidistant
            eq_pos = torch.linspace(0, 1, max_n_q, device=self.device, dtype=self.dtype).expand(half, max_n_q)
            positions[:half] = eq_pos

            rand_pos = torch.rand(batch_size - half, max_n_q, device=self.device, dtype=self.dtype) # other half gets sorted random
            rand_pos, _ = rand_pos.sort(dim=-1)
            positions[half:] = rand_pos
        else:
            raise ValueError(f"Unknown spacing mode: {self.mode}")

        q = q_min[:, None] + positions * (q_max - q_min)[:, None]

        n_qs = torch.randint(self.n_q_range[0], self.n_q_range[1] + 1, (batch_size,), device=self.device)

        if 'params' in context and self.total_thickness_constraint: ### N_points > 1 + (Q_spread * total_thickness * min_np_per_kiessing_fringe) / (2*pi)
            d_total = context['params'].thicknesses.sum(-1)
            limit = 1 + ((q_max - q_min) * d_total * self.min_points_per_fringe) / (2*np.pi)
            limit = limit.ceil().int()
            n_qs = torch.maximum(n_qs, limit)
            n_qs = torch.clamp(n_qs, max=self.n_q_range[1])

        indices = torch.arange(max_n_q, device=self.device).expand(batch_size, max_n_q)
        valid_mask = indices < n_qs[:, None] # right side padding
        
        if self.shuffle_mask: # shuffle valid positions (inter-spread padding)
            perm = torch.argsort(torch.rand(batch_size, max_n_q, device=self.device), dim=-1)
            valid_mask = torch.gather(valid_mask, dim=1, index=perm)

        context['key_padding_mask'] = valid_mask
        context['n_points'] = valid_mask.sum(dim=-1)

        return q
    
    def scale_q(self, q):
        scaled_q_01 = (q - self.q_min_range[0]) / (self.q_max_range[1] - self.q_min_range[0]) 

        return 2.0 * (scaled_q_01 - 0.5)



# -*- coding: utf-8 -*-

import torch
from torch import Tensor


def kinematical_approximation(
        q: Tensor,
        thickness: Tensor,
        roughness: Tensor,
        sld: Tensor,
        *,
        apply_fresnel: bool = True,
        log: bool = False,
):
    """Simulates reflectivity curves for SLD profiles with box model parameterization using the kinematical approximation

    Args:
        q (Tensor): tensor of momentum transfer (q) values with shape [batch_size, n_points] or [n_points]
        thickness (Tensor): tensor containing the layer thicknesses (ordered from top to bottom) with shape [batch_size, n_layers]
        roughness (Tensor): tensor containing the interlayer roughnesses (ordered from top to bottom) with shape [batch_size, n_layers + 1]
        sld (Tensor): tensor containing the layer SLDs (real or complex; ordered from top to bottom) with shape [batch_size, n_layers + 1]. 
                    It includes the substrate but excludes the ambient medium which is assumed to have an SLD of 0.
        apply_fresnel (bool, optional): whether to use the Fresnel coefficient in the computation. Defaults to ``True``.
        log (bool, optional): if True the base 10 logarithm of the reflectivity curves is returned. Defaults to ``False``.

    Returns:
        Tensor: tensor containing the simulated reflectivity curves with shape [batch_size, n_points]
    """
    c_dtype = torch.complex128 if q.dtype is torch.float64 else torch.complex64

    batch_size, num_layers = thickness.shape

    q = q.to(c_dtype)

    if q.dim() == 1:
        q.unsqueeze_(0)

    if q.dim() == 2:
        q.unsqueeze_(-1)

    sld = sld * 1e-6 + 1e-30j

    drho = torch.cat([sld[..., 0][..., None], sld[..., 1:] - sld[..., :-1]], -1)[:, None]
    thickness = torch.cumsum(torch.cat([torch.zeros(batch_size, 1).to(thickness), thickness], -1), -1)[:, None]
    roughness = roughness[:, None]

    r = (drho * torch.exp(- (roughness * q) ** 2 / 2 + 1j * (q * thickness))).sum(-1).abs().float() ** 2

    if apply_fresnel:

        substrate_sld = sld[:, -1:]

        rf = _get_fresnel_reflectivity(q, substrate_sld[:, None])

        r = torch.clamp_max_(r * rf / substrate_sld.real ** 2, 1.)

    if log:
        r = torch.log10(r)

    return r


def _get_fresnel_reflectivity(q, substrate_slds):
    _RE_CONST = 0.28174103675406496 # 2/sqrt(16*pi)

    q_c = torch.sqrt(substrate_slds + 0j) / _RE_CONST * 2
    q_prime = torch.sqrt(q ** 2 - q_c ** 2 + 0j)
    r_f = ((q - q_prime) / (q + q_prime)).abs().float() ** 2

    return r_f.squeeze(-1)


import numpy as np
from numpy import ndarray

from torch import Tensor, tensor

__all__ = [
    'to_np',
    'to_t',
    'angle_to_q',
    'q_to_angle',
    'energy_to_wavelength',
    'wavelength_to_energy',
]


def to_np(arr):
    """Converts Pytorch tensor or Python list to Numpy array

    Args:
        arr (torch.Tensor or list): Input Pytorch tensor or Python list

    Returns:
        numpy.ndarray: Converted Numpy array
    """
    
    if isinstance(arr, Tensor):
        return arr.detach().cpu().numpy()
    return np.asarray(arr)


def to_t(arr, device=None, dtype=None):
    """Converts Numpy array or Python list to Pytorch tensor

    Args:
        arr (numpy.ndarray or list): Input 
        device (torch.device or str, optional): device for the tensor ('cpu', 'cuda')
        dtype (torch.dtype, optional): data type of the tensor (e.g. torch.float32)

    Returns:
        torch.Tensor: converted Pytorch tensor
    """
    
    if not isinstance(arr, Tensor):
        return tensor(arr, device=device, dtype=dtype)
    return arr


# taken from mlreflect package
# mlreflect/xrrloader/dataloader/transform.py

def angle_to_q(scattering_angle: ndarray or float, wavelength: float):
    """Conversion from full scattering angle (degrees) to scattering vector (inverse angstroms)"""
    return 4 * np.pi / wavelength * np.sin(scattering_angle / 2 * np.pi / 180)


def q_to_angle(q: ndarray or float, wavelength: float):
    """Conversion from scattering vector (inverse angstroms) to full scattering angle (degrees)"""
    return 2 * np.arcsin(q * wavelength / (4 * np.pi)) / np.pi * 180


def energy_to_wavelength(energy: float):
    """Conversion from photon energy (eV) to photon wavelength (angstroms)"""
    return 1.2398 / energy * 1e4


def wavelength_to_energy(wavelength: float):
    """Conversion from photon wavelength (angstroms) to photon energy (eV)"""
    return 1.2398 / wavelength * 1e4

def get_filtering_mask(Q, R, dR, threshold=0.3, consecutive=3,
                        remove_singles=True, remove_consecutives=True,
                        q_start_trunc=0.1):
    Q, R, dR = Q.copy(), R.copy(), dR.copy()
    rel_error = np.abs(dR / R)

    # Mask for singles
    mask_singles = (rel_error >= threshold) if remove_singles else np.zeros_like(Q, dtype=bool)

    # Mask for truncation
    mask_consecutive = np.zeros_like(Q, dtype=bool)
    if remove_consecutives:
        count = 0
        cutoff_idx = None
        for i in range(len(Q)):
            if Q[i] < q_start_trunc:
                continue
            if rel_error[i] >= threshold:
                count += 1
                if count >= consecutive:
                    cutoff_idx = i - consecutive + 1
                    break
            else:
                count = 0
        if cutoff_idx is not None:
            mask_consecutive[cutoff_idx:] = True

    final_mask = mask_singles | mask_consecutive
    return ~final_mask


from typing import Tuple, Union, List

import torch
from torch import Tensor

from reflectorch.data_generation.utils import (
    uniform_sampler,
    logdist_sampler,
)

from reflectorch.data_generation.priors.scaler_mixin import ScalerMixin
from reflectorch.data_generation.priors.base import PriorSampler
from reflectorch.data_generation.priors.subprior_sampler import UniformSubPriorParams
from reflectorch.data_generation.priors.params import Params
from reflectorch.data_generation.priors.utils import get_max_allowed_roughness
from reflectorch.data_generation.priors.no_constraints import (
    DEFAULT_DEVICE,
    DEFAULT_DTYPE,
)


class ExpUniformSubPriorSampler(PriorSampler, ScalerMixin):
    PARAM_CLS = UniformSubPriorParams

    def __init__(self,
                 params: List[Union[float, Tuple[float, float], Tuple[float, float, float, float]]],
                 device: torch.device = DEFAULT_DEVICE,
                 dtype: torch.dtype = DEFAULT_DTYPE,
                 scaled_range: Tuple[float, float] = (-1, 1),
                 logdist: bool = False,
                 relative_min_bound_width: float = 1e-4,
                 smaller_roughnesses: bool = True,
                 ):
        self.device = device
        self.dtype = dtype
        self.scaled_range = scaled_range
        self.relative_min_bound_width = relative_min_bound_width
        self.logdist = logdist
        self.smaller_roughnesses = smaller_roughnesses
        self._init_params(*params)

    @property
    def max_num_layers(self) -> int:
        return self.num_layers

    def _init_params(self, *params: Union[float, Tuple[float, float], Tuple[float, float, float, float]]):
        self.num_layers = (len(params) - 2) // 3
        self._total_num_params = len(params)

        fixed_mask = []
        bounds = []
        delta_bounds = []
        param_dim = 0

        for param in params:
            if isinstance(param, (float, int)):
                deltas = (0, 0)
                param = (param, param)
                fixed_mask.append(True)
            else:
                param_dim += 1
                fixed_mask.append(False)

                if len(param) == 4:
                    param, deltas = param[:2], param[2:]
                else:
                    max_delta = param[1] - param[0]
                    deltas = (max_delta * self.relative_min_bound_width, max_delta)

            bounds.append(param)
            delta_bounds.append(deltas)

        self.fixed_mask = torch.tensor(fixed_mask).to(self.device)
        self.fitted_mask = ~self.fixed_mask
        self.roughnesses_mask = torch.zeros_like(self.fitted_mask)
        self.roughnesses_mask[self.num_layers: self.num_layers * 2 + 1] = True

        self.min_bounds, self.max_bounds = torch.tensor(bounds).to(self.device).to(self.dtype).T
        self.min_deltas, self.max_deltas = map(torch.atleast_2d, torch.tensor(delta_bounds).to(self.min_bounds).T)
        self._param_dim = param_dim
        self._num_fixed = self.fixed_mask.sum()

        self.fixed_params = self.min_bounds[self.fixed_mask]

    @property
    def param_dim(self) -> int:
        return self._param_dim

    def sample(self, batch_size: int) -> UniformSubPriorParams:
        min_bounds, max_bounds = self.sample_bounds(batch_size)

        params = torch.rand(
            *min_bounds.shape,
            device=self.device,
            dtype=self.dtype
        ) * (max_bounds - min_bounds) + min_bounds

        thicknesses, roughnesses, slds = torch.split(
            params, [self.max_num_layers, self.max_num_layers + 1, self.max_num_layers + 1], dim=-1
        )

        if self.smaller_roughnesses:
            fitted_r_mask = self.fitted_mask.clone()
            fitted_r_mask[~self.roughnesses_mask] = False

            min_roughness = self.min_bounds[fitted_r_mask]
            max_roughness = torch.clamp(
                get_max_allowed_roughness(thicknesses)[..., self.fitted_mask[self.roughnesses_mask]],
                min_roughness,
                self.max_bounds[fitted_r_mask]
            )

            min_vector = self.min_bounds.clone()[None].repeat(batch_size, 1)
            max_vector = self.max_bounds.clone()[None].repeat(batch_size, 1)

            max_vector[..., fitted_r_mask] = max_roughness

            assert torch.all(max_vector[..., fitted_r_mask] == max_roughness)

            min_deltas = self.min_deltas.clone().repeat(batch_size, 1)
            max_deltas = self.max_deltas.clone().repeat(batch_size, 1)

            max_deltas[..., fitted_r_mask] = torch.clamp_max(
                max_deltas[..., fitted_r_mask],
                max_roughness - min_roughness,
            )

            fitted_mask = torch.zeros_like(self.fitted_mask)
            fitted_mask[fitted_r_mask] = True

            updated_min_bounds, updated_max_bounds = self._sample_bounds(
                batch_size, min_vector, max_vector, min_deltas, max_deltas, fitted_mask
            )

            min_bounds[..., fitted_mask], max_bounds[..., fitted_mask] = (
                updated_min_bounds[..., fitted_mask], updated_max_bounds[..., fitted_mask]
            )

            params[..., fitted_mask] = torch.rand(
                batch_size, fitted_mask.sum().item(),
                device=self.device,
                dtype=self.dtype
            ) * (max_bounds[..., fitted_mask] - min_bounds[..., fitted_mask]) + min_bounds[..., fitted_mask]

            thicknesses, roughnesses, slds = torch.split(
                params, [self.max_num_layers, self.max_num_layers + 1, self.max_num_layers + 1], dim=-1
            )

        params = UniformSubPriorParams(thicknesses, roughnesses, slds, min_bounds, max_bounds)

        return params

    def scale_params(self, params: UniformSubPriorParams) -> Tensor:
        params_t = params.as_tensor(add_bounds=False)

        scaled_params = self._scale(params_t, params.min_bounds, params.max_bounds)[..., self.fitted_mask]

        scaled_min_bounds = self._scale(params.min_bounds, self.min_bounds, self.max_bounds)[..., self.fitted_mask]

        scaled_max_bounds = self._scale(params.max_bounds, self.min_bounds, self.max_bounds)[..., self.fitted_mask]

        scaled_params = torch.cat([scaled_params, scaled_min_bounds, scaled_max_bounds], -1)

        return scaled_params

    def scale_bounds(self, bounds: Tensor) -> Tensor:
        return self._scale(bounds, self.min_bounds, self.max_bounds)[..., self.fitted_mask]

    def restore_params(self, scaled_params: Tensor) -> UniformSubPriorParams:
        scaled_params, scaled_min_bounds, scaled_max_bounds = torch.split(
            scaled_params, [self.param_dim, self.param_dim, self.param_dim], dim=1
        )

        min_bounds = self._restore(
            scaled_min_bounds, self.min_bounds[self.fitted_mask], self.max_bounds[self.fitted_mask]
        )
        max_bounds = self._restore(
            scaled_max_bounds, self.min_bounds[self.fitted_mask], self.max_bounds[self.fitted_mask]
        )

        restored_params = self._restore(scaled_params, min_bounds, max_bounds)

        params_t = torch.cat(
            [
                self._cat_restored_with_fixed_vector(restored_params),
                self._cat_restored_with_fixed_vector(min_bounds),
                self._cat_restored_with_fixed_vector(max_bounds),
            ], -1
        )

        params = UniformSubPriorParams.from_tensor(params_t)

        return params

    def _cat_restored_with_fixed_vector(self, restored_t: Tensor) -> Tensor:
        return self._cat_fitted_fixed_t(restored_t, self.fixed_params)

    def _cat_fitted_fixed_t(self, fitted_t: Tensor, fixed_t: Tensor, fitted_mask: Tensor = None) -> Tensor:
        if fitted_mask is None:
            fitted_mask = self.fitted_mask
            fixed_mask = self.fixed_mask
        else:
            fixed_mask = ~fitted_mask

        total_num_params = self.fitted_mask.sum().item() + self.fixed_mask.sum().item()

        batch_size = fitted_t.shape[0]

        concat_t = torch.empty(
            batch_size, total_num_params, device=fitted_t.device, dtype=fitted_t.dtype
        )
        concat_t[:, fitted_mask] = fitted_t
        concat_t[:, fixed_mask] = fixed_t[None].expand(batch_size, -1)

        return concat_t

    def log_prob(self, params: UniformSubPriorParams) -> Tensor:
        log_prob = torch.zeros(params.batch_size, device=params.device, dtype=params.dtype)
        indices = self.get_indices_within_bounds(params)
        log_prob[~indices] = float('-inf')
        return log_prob

    def get_indices_within_bounds(self, params: UniformSubPriorParams) -> Tensor:
        t_params = torch.cat([
            params.thicknesses,
            params.roughnesses,
            params.slds
        ], dim=-1)

        indices = (
                torch.all(t_params >= params.min_bounds, dim=-1) &
                torch.all(t_params <= params.max_bounds, dim=-1)
        )

        return indices

    def clamp_params(self, params: UniformSubPriorParams) -> UniformSubPriorParams:
        params = UniformSubPriorParams.from_tensor(
            torch.cat([
                torch.clamp(
                    params.as_tensor(add_bounds=False),
                    params.min_bounds, params.max_bounds
                ),
                params.min_bounds, params.max_bounds
            ], dim=1)
        )
        return params

    def get_indices_within_domain(self, params: UniformSubPriorParams) -> Tensor:
        return self.get_indices_within_bounds(params)

    def sample_bounds(self, batch_size: int):
        return self._sample_bounds(
            batch_size,
            self.min_bounds,
            self.max_bounds,
            self.min_deltas,
            self.max_deltas,
            self.fitted_mask,
        )

    def _sample_bounds(self, batch_size, min_vector, max_vector, min_deltas, max_deltas, fitted_mask):
        if self.logdist:
            widths_sampler_func = logdist_sampler
        else:
            widths_sampler_func = uniform_sampler

        num_fitted = fitted_mask.sum().item()
        num_fixed = fitted_mask.numel() - num_fitted

        prior_widths = widths_sampler_func(
            min_deltas[..., fitted_mask], max_deltas[..., fitted_mask],
            batch_size, num_fitted,
            device=self.device, dtype=self.dtype
        )

        prior_widths = self._cat_fitted_fixed_t(prior_widths, torch.zeros(num_fixed).to(prior_widths), fitted_mask)

        prior_centers = uniform_sampler(
            min_vector + prior_widths / 2, max_vector - prior_widths / 2,
            *prior_widths.shape,
            device=self.device, dtype=self.dtype
        )

        min_bounds, max_bounds = prior_centers - prior_widths / 2, prior_centers + prior_widths / 2

        return min_bounds, max_bounds

    @staticmethod
    def scale_bounds_with_q(bounds: Tensor, q_ratio: float) -> Tensor:
        params = Params.from_tensor(torch.atleast_2d(bounds).clone())
        params.scale_with_q(q_ratio)
        return params.as_tensor().squeeze()

    def clamp_bounds(self, bounds: Tensor) -> Tensor:
        return torch.clamp(
            torch.atleast_2d(bounds), torch.atleast_2d(self.min_bounds), torch.atleast_2d(self.max_bounds)
        )



from typing import Any

__all__ = [
    "ProcessData",
    "ProcessPipeline",
]


class ProcessData(object):
    def __add__(self, other):
        if isinstance(other, ProcessData):
            return ProcessPipeline(self, other)

    def apply(self, args: Any, context: dict = None):
        return args

    def __call__(self, args: Any, context: dict = None):
        return self.apply(args, context)

    def __repr__(self):
        return f'{self.__class__.__name__}()'


class ProcessPipeline(ProcessData):
    def __init__(self, *processes):
        self._processes = list(processes)

    def apply(self, args: Any, context: dict = None):
        for process in self._processes:
            args = process(args, context)
        return args

    def __add__(self, other):
        if isinstance(other, ProcessPipeline):
            return ProcessPipeline(*self._processes, *other._processes)
        elif isinstance(other, ProcessData):
            return ProcessPipeline(*self._processes, other)

    def __repr__(self):
        processes = ", ".join(repr(p) for p in self._processes)
        return f'ProcessPipeline({processes})'



from pathlib import Path

import torch
from torch import Tensor

from reflectorch.data_generation.priors import PriorSampler
from reflectorch.paths import SAVED_MODELS_DIR


class CurvesScaler(object):
    """Base class for curve scalers"""
    def scale(self, curves: Tensor):
        raise NotImplementedError

    def restore(self, curves: Tensor):
        raise NotImplementedError


class LogAffineCurvesScaler(CurvesScaler):
    """    Curve scaler which scales the reflectivity curves according to the logarithmic affine transformation:
    :math:`\log_{10}(R + eps) \cdot weight + bias`.

    Args:
        weight (float): multiplication factor in the transformation
        bias (float): addition term in the transformation
        eps (float): sets the minimum intensity value of the reflectivity curves which is considered
    """
    def __init__(self, weight: float = 0.1, bias: float = 0.5, eps: float = 1e-10):
        self.weight = weight
        self.bias = bias
        self.eps = eps

    def scale(self, curves: Tensor):
        """scales the reflectivity curves to a ML-friendly range

        Args:
            curves (Tensor): original reflectivity curves

        Returns:
            Tensor: reflectivity curves scaled to a ML-friendly range
        """
        return torch.log10(curves + self.eps) * self.weight + self.bias

    def restore(self, curves: Tensor):
        """restores the physical reflectivity curves

        Args:
            curves (Tensor): scaled reflectivity curves

        Returns:
            Tensor: reflectivity curves restored to the physical range
        """
        return 10 ** ((curves - self.bias) / self.weight) - self.eps


class MeanNormalizationCurvesScaler(CurvesScaler):
    """Curve scaler which scales the reflectivity curves by the precomputed mean of a batch of curves

    Args:
        path (str, optional): path to the precomputed mean of the curves, only used if ``curves_mean`` is None. Defaults to None.
        curves_mean (Tensor, optional): the precomputed mean of the curves. Defaults to None.
        device (torch.device, optional): the Pytorch device. Defaults to 'cuda'.
    """

    def __init__(self, path: str = None, curves_mean: Tensor = None, device: torch.device = 'cuda'):
        if curves_mean is None:
            curves_mean = torch.load(self.get_path(path))
        self.curves_mean = curves_mean.to(device)

    def scale(self, curves: Tensor):
        """scales the reflectivity curves to a ML-friendly range

        Args:
            curves (Tensor): original reflectivity curves

        Returns:
            Tensor: reflectivity curves scaled to a ML-friendly range
        """
        self.curves_mean = self.curves_mean.to(curves)
        return curves / self.curves_mean - 1

    def restore(self, curves: Tensor):
        """restores the physical reflectivity curves

        Args:
            curves (Tensor): scaled reflectivity curves

        Returns:
            Tensor: reflectivity curves restored to the physical range
        """
        self.curves_mean = self.curves_mean.to(curves)
        return (curves + 1) * self.curves_mean

    @staticmethod
    def save(prior_sampler: PriorSampler, q: Tensor, path: str, num: int = 16384):
        """computes the mean of a batch of reflectivity curves and saves it

        Args:
            prior_sampler (PriorSampler): the prior sampler
            q (Tensor): the q values
            path (str): the path for saving the mean of the curves
            num (int, optional): the number of curves used to compute the mean. Defaults to 16384.
        """
        params = prior_sampler.sample(num)
        curves_mean = params.reflectivity(q, log=False).mean(0).cpu()
        torch.save(curves_mean, MeanNormalizationCurvesScaler.get_path(path))

    @staticmethod
    def get_path(path: str) -> Path:
        if not path.endswith('.pt'):
            path = path + '.pt'
        return SAVED_MODELS_DIR / path



from reflectorch.data_generation.priors.params import Params
from reflectorch.data_generation.priors.base import PriorSampler
from reflectorch.data_generation.priors.no_constraints import BasicPriorSampler
from reflectorch.data_generation.priors.independent_priors import (
    SingleParamPrior,
    SimplePriorSampler,
    UniformParamPrior,
    GaussianParamPrior,
    TruncatedGaussianParamPrior
)

from reflectorch.data_generation.priors.subprior_sampler import (
    UniformSubPriorParams,
    UniformSubPriorSampler,
    NarrowSldUniformSubPriorSampler,
)
from reflectorch.data_generation.priors.exp_subprior_sampler import ExpUniformSubPriorSampler
from reflectorch.data_generation.priors.multilayer_structures import (
    SimpleMultilayerSampler,
    MultilayerStructureParams,
)
from reflectorch.data_generation.priors.parametric_models import (
    ParametricModel,
    MULTILAYER_MODELS,
)
from reflectorch.data_generation.priors.parametric_subpriors import (
    SubpriorParametricSampler,
    BasicParams,
)
from reflectorch.data_generation.priors.sampler_strategies import (
    SamplerStrategy,
    BasicSamplerStrategy,
    ConstrainedRoughnessSamplerStrategy,
    ConstrainedRoughnessAndImgSldSamplerStrategy,
)

__all__ = [
    "SingleParamPrior",
    "SimplePriorSampler",
    "UniformParamPrior",
    "GaussianParamPrior",
    "TruncatedGaussianParamPrior",
    "Params",
    "PriorSampler",
    "BasicPriorSampler",
    "UniformSubPriorParams",
    "UniformSubPriorSampler",
    "NarrowSldUniformSubPriorSampler",
    "ExpUniformSubPriorSampler",
    "SimpleMultilayerSampler",
    "MultilayerStructureParams",
    "SubpriorParametricSampler",
    "BasicParams",
    "ParametricModel",
    "MULTILAYER_MODELS",
    "SamplerStrategy",
    "BasicSamplerStrategy",
    "ConstrainedRoughnessSamplerStrategy",
    "ConstrainedRoughnessAndImgSldSamplerStrategy",
]



import pytest
import torch
from reflectorch import ResidualMLP


@pytest.mark.parametrize("dim_in", [256])
@pytest.mark.parametrize("dim_out", [10])
@pytest.mark.parametrize("residual", [False, True])
@pytest.mark.parametrize("repeats_per_block", [1, 2])
@pytest.mark.parametrize("activation", ['relu', 'rowdy'])
@pytest.mark.parametrize("dropout_rate", [0, 0.2])
def test_mlp(dim_in, dim_out, residual, repeats_per_block, activation, dropout_rate):
    adaptive_activation = True if activation == 'rowdy' else False

    mlp = ResidualMLP(
        dim_in = dim_in,
        dim_out = dim_out,
        dim_condition = 0,
        layer_width = 128,
        num_blocks = 4,
        repeats_per_block= repeats_per_block,
        activation = activation,
        use_batch_norm = True,
        dropout_rate = dropout_rate,
        residual = residual,
        adaptive_activation = adaptive_activation,
        conditioning = 'concat',
    )

    batch_size = 8
    input_tensor = torch.randn((batch_size, dim_in))

    assert mlp(input_tensor).shape == (batch_size, dim_out)

